#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 23 10:46:42 2022

@author: alex
"""

##############################################################################
import numpy as np
##############################################################################
from pyro.dynamic import system
##############################################################################

import matplotlib.pyplot as plt



##############################################################################
# 1 DoF Car Model
##############################################################################
        
class LongitudinalFrontWheelDriveCarWithWheelSlipInput( system.ContinuousDynamicSystem ):
    """ 
    Equations of Motion
    -------------------------
    TBD
    """
    
    ############################
    def __init__(self):
        """ """
    
        # Dimensions
        self.n = 2   
        self.m = 1   
        self.p = 2
        
        # initialize standard params
        system.ContinuousDynamicSystem.__init__( self, self.n, self.m, self.p)
        
        # Labels
        self.name = 'Front Wheel Drive Car'
        self.state_label = ['x','dx']
        self.input_label = ['slip']
        self.output_label = ['x','dx']
        
        # Units
        self.state_units = ['[m]','[m/sec]']
        self.input_units = ['[]']
        self.output_units = ['[m]','[m/sec]']
        
        # State working range
        self.x_ub = np.array([+50,+30,])
        self.x_lb = np.array([ 0,-10])
        
        # Input working range
        self.u_ub = np.array([  0.3])
        self.u_lb = np.array([ -0.3])
        
        # Model param
        self.lenght  = 2          # distance between front wheel and back wheel [m]
        self.xc      = 1          # distance from back wheel to c.g.  [m]
        self.yc      = 0.5        # height from ground to c.g.  [m]
        self.mass    = 1500       # total car mass [kg]
        self.gravity = 9.81       # gravity constant [N/kg]
        self.rho     = 1.225      # air density [kg/m3]
        self.cdA     = 0.3 * 2    # drag coef time area [m2]
        
        # Ground traction curve parameters
        self.mu_max   = 1.0
        self.mu_slope = 70.
        
        
        # Graphic output parameters 
        self.dynamic_domain  = False
        self.dynamic_range   = self.lenght * 2
        
        # Animation output graphical parameters
        self.linestyle = '-'
        self.obs_dist =  self.x_ub[0] + self.lenght * 2 # using the upper bound on x range
        self.obs_size = 2
        
        
    #############################
    def compute_ratios(self):
        """ Shorcut function for comuting usefull length ratios """
        
        ry = self.yc / self.lenght
        rr = self.xc / self.lenght # ratio of space rear of the c.g.
        rf = 1 - rr                # ratio of space in front of c.g.
        
        return ry, rr, rf
    
    #############################
    def slip2force(self, slip ):
        """ Shorcut function for comuting usefull length ratios """
        
        # Sigmoid
        mu = self.mu_max * ( 2 / ( 1 + np.exp( -self.mu_slope * slip ) ) - 1 )
        
        return mu
    
    
    #############################
    def plot_slip2force(self, slip_min = -0.2, slip_max = 0.2 ):
        """ Shorcut function for comuting usefull length ratios """
        
        slips = np.arange( slip_min, slip_max, 0.001 )
        mus   = self.slip2force( slips )
        
        fig = plt.figure(figsize=(4, 2), dpi=300, frameon=True)
        fig.canvas.manager.set_window_title('Ground traction curve')
        ax  = fig.add_subplot(1, 1, 1)
        
        ax.plot( slips , mus , 'b')
        ax.set_ylabel('mu = |Fx/Fz|', fontsize=5)
        ax.set_xlabel('Slip ratio', fontsize=5 )
        ax.tick_params( labelsize = 5 )
        ax.grid(True)
        
        fig.tight_layout()
        fig.canvas.draw()
        
        plt.show()
    
    
        
    #############################
    def f(self, x , u , t = 0 ):
        """ 
        Continuous time foward dynamics evaluation
        
        dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUPUTS
        dx : state derivative vectror n x 1
        
        """
        
        dx = np.zeros(self.n) # State derivative vector
        
        ###################
        
        slip = u
        v    = x[1]
        
        # compute ratio of horizontal/vertical force
        mu = self.slip2force( slip ) 
        
        # constant params local vairables
        ry, rr, rf = self.compute_ratios() 
        m    = self.mass 
        g    = self.gravity
        rcda = self.rho * self.cdA
        
        # Drag froce
        fd = 0.5 * rcda * v * np.abs( v ) # drag froce with the right sign
        
        # Acceleration (equation considering weight transfer)
        a  = (mu * m * g * rr - fd )/( m * (1 + mu * ry ))
        
        ###################
        
        dx[0]  = v # velocity
        dx[1]  = a # acc
        
        # ###################
        # # Normal force check
        # fn_front = m * g * rr - m * a * ry
        # fn_rear  = m * g * rf + m * a * ry
        # if (fn_front<0) :
        #     print('Normal force on front wheel is negative: fn = ', fn_front)
        # if (fn_rear<0) : 
        #     print('Normal force on rear wheel is negative: fn = ', fn_rear)
        # ###################
        
        return dx
    
    #############################
    def isavalidinput(self , x , u):
        """ check if u is in the control inputs domain given x """
        
        ans = False
        
        # Min-Max Slip
        
        for i in range(self.m):
            ans = ans or ( u[i] < self.u_lb[i] )
            ans = ans or ( u[i] > self.u_ub[i] )
        
        # Normal Forces Negatives?
        
        slip = u
        v    = x[1]
        
        # compute ratio of horizontal/vertical force
        mu = self.slip2force( slip ) 
        
        # constant params local vairables
        ry, rr, rf = self.compute_ratios() 
        m          = self.mass 
        g          = self.gravity
        rcda       = self.rho * self.cdA
        
        # Drag froce
        fd = 0.5 * rcda * v * np.abs( v ) # drag froce with the right sign
        
        # Acceleration (equation considering weight transfer)
        a  = (mu * m * g * rr - fd )/( m * (1 + mu * ry ))
        
        # Normal force check
        fn_front = m * g * rr - m * a * ry
        fn_rear  = m * g * rf + m * a * ry
        
        ans = ans or ( fn_front < 0. )
        ans = ans or ( fn_rear  < 0. )
            
        return not(ans)
    
    
    ###########################################################################
    # For graphical output
    ###########################################################################
    
    
    #############################
    def xut2q( self, x , u , t ):
        """ compute config q """
        
        q   = np.append(  x , u[0] ) # steering angle is part of the config
        
        return q
    
    
    ###########################################################################
    def forward_kinematic_domain(self, q ):
        """ 
        """
        l = self.dynamic_range
        
        x = q[0]
        y = 0
        z = 0
        
        if self.dynamic_domain:
        
            domain  = [ ( -l + x , l + x ) ,
                        ( -l + y , l + y ) ,
                        ( -l + z , l + z ) ]#  
        else:
            
            domain  = [ ( 0 , self.obs_dist + self.obs_size * 2 ) ,
                        ( 0 , 1 ) ,
                        ( 0 , self.obs_dist + self.obs_size * 2 ) ]#
            
                
        return domain
    
    
    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """ 
        Compute points p = [x;y;z] positions given config q 
        ----------------------------------------------------
        - points of interest for ploting
        
        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines
        
        """
        
        # Variables
        
        travel   = q[0]
        slipping = (np.abs( q[2] ) > 0.03 ) # bool
        
        
        
        lines_pts = [] # list of array (n_pts x 3) for each lines
        
        
        ###########################
        # bottom line
        ###########################
        
        pts = np.zeros((2,3))
        
        pts[0,:] = [ -1000  , 0 , 0 ]
        pts[1,:] = [  1000  , 0 , 0 ]
        
        lines_pts.append( pts )
        
        ###########################
        # obstacle
        ###########################
        
        pts = np.zeros((5,3))
        
        d = self.obs_size
        
        pts[0,:] = [ 0  , 0 , 0 ]
        pts[1,:] = [ d  , 0 , 0 ]
        pts[2,:] = [ d  , d , 0 ]
        pts[3,:] = [ 0  , d , 0 ]
        pts[4,:] = [ 0  , 0 , 0 ]
        
        pts[:,0] = pts[:,0] + self.obs_dist

        
        lines_pts.append( pts )
        
        ###########################
        # Vehicule
        ###########################
        
        pts = np.zeros((13,3))
        
        r = 0.3
        x1 = 1
        y1 = 1
        y2 = 1.5
        y3 = 1.4
        x2 = 1
        x3 = 1
        y3 = 0.6
        
        l = self.lenght
        
        pts[0,:]  = [ 0  , 0 , 0 ]
        pts[1,:]  = [ -x1  , 0 , 0 ]
        pts[2,:]  = [ -x1  , y1 , 0 ]
        pts[3,:]  = [ 0  , y2 , 0 ]
        pts[4,:]  = [ l , y2 , 0 ]
        pts[5,:]  = [ l - x2 , y2 , 0 ]
        pts[6,:]  = [ l - x2  , y1 , 0 ]
        pts[7,:]  = [ l  , y1 , 0 ]
        pts[8,:]  = [ l  , y2 , 0 ]
        pts[9,:]  = [ l  , y1 , 0 ]
        pts[10,:] = [ l+x3  , y3 , 0 ]
        pts[11,:] = [ l+x3  , 0 , 0 ]
        pts[12,:] = [ 0  , 0 , 0 ]


        pts[:,0] = pts[:,0] + travel  # translate horizontally the car postion
        pts[:,1] = pts[:,1] + r       # translate vertically the wheel radius
        
        lines_pts.append( pts )
        
        ###########################
        # Wheels
        ###########################
        
        if slipping:
            r = r*1.2
        
        angles = np.arange(0,6.4,0.1)
        n      = angles.size
        
        pts = np.zeros((n,3))
        
        for i in range(n):
            a = angles[i]
            pts[i,:] = [ r * np.cos(a) , r * np.sin(a) , 0 ]

        pts[:,0] = pts[:,0] + travel
        pts[:,1] = pts[:,1] + r
        
        lines_pts.append( pts )
        
        pts = pts.copy()
        pts[:,0] = pts[:,0] + l
        
        lines_pts.append( pts )
                
        return lines_pts
    
    

##############################################################################
# 2 DoF Car Model
##############################################################################
        
class LongitudinalFrontWheelDriveCarWithTorqueInput( LongitudinalFrontWheelDriveCarWithWheelSlipInput ):
    """ 
    Equations of Motion
    -------------------------
    TBD
    """
    
    ############################
    def __init__(self):
        """ """
        
        # initialize standard params
        super().__init__()
        
        # Dimensions
        self.n = 4   
        self.m = 1   
        self.p = 1
        
        # initialize standard vectors with new dimensions
        system.ContinuousDynamicSystem.__init__(self,self.n,self.m,self.p)

        # Labels
        self.name = 'Front Wheel Drive Car'
        self.state_label = ['x','dx', 'w','theta']
        self.input_label = ['torque']
        self.output_label = ['slip']
        
        # Units
        self.state_units = ['[m]','[m/sec]','[rad/sec]','[rad]']
        self.input_units = ['[Nm]']
        self.output_units = ['']
        
        # State working range
        self.x_ub = np.array([+50,+30,1000, 10000])
        self.x_lb = np.array([ 0,-10,-100,-10000])
        
        # Input working range
        self.u_ub = np.array([  200])
        self.u_lb = np.array([ -200])
        
        # Additionnal Model param
        self.wheel_radius  = 0.3 # [m]
        self.wheel_inertia = 1.5 # [kg m2]
        
        self.x0 = np.array([0,0.01,0,0])
        
        self.linestyle = '-'
        self.dynamic_domain = True
        
        
    #############################
    def f(self, x , u , t = 0 ):
        """ 
        Continuous time foward dynamics evaluation
        
        dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUPUTS
        dx : state derivative vectror n x 1
        
        """
        
        dx = np.zeros(self.n) # State derivative vector
        
        ###################
        
        torque = u
        v      = x[1]
        w      = x[2]
        
        # constant params local vairables
        ry, rr, rf = self.compute_ratios() 
        m    = self.mass 
        g    = self.gravity
        rcda = self.rho * self.cdA
        r    = self.wheel_radius
        j    = self.wheel_inertia
        
        #slip computation
        slip = np.clip( ( r * w - v ) / (np.abs(v) + 0.0 ) , -0.5 , 0.5 )
        
        # compute ratio of horizontal/vertical force
        mu = self.slip2force( slip ) 
        
        # Drag froce
        fd = 0.5 * rcda * v * np.abs( v ) # drag froce with the right sign
        
        # Acceleration (equation considering weight transfer)
        a  = (mu * m * g * rr - fd )/( m * (1 + mu * ry ))
        
        # Wheel acceleration
        dw = (torque - r * (m * a + fd)) / j
        
        ###################
        
        dx[0]  = v   # velocity
        dx[1]  = a   # acc
        dx[2]  = dw  # angular acc. of the wheels
        dx[3]  = w   #
        
        ###################
        # Normal force check
        fn_front = m * g * rr - m * a * ry
        fn_rear  = m * g * rf + m * a * ry
        if (fn_front<0) :
            print('Normal force on front wheel is negative: fn = ', fn_front)
        if (fn_rear<0) : 
            print('Normal force on rear wheel is negative: fn = ', fn_rear)
        ###################
        
        return dx
    
    
    #############################
    def h(self, x , u , t = 0 ):
        """ 
        Continuous time foward dynamics evaluation
        
        dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUPUTS
        dx : state derivative vectror n x 1
        
        """
        
        y = np.zeros(self.p) # State derivative vector
        
        ###################
        v      = x[1]
        w      = x[2]
        
        # constant params local vairables
        r    = self.wheel_radius
        
        #slip computation
        slip = np.clip( ( r * w - v ) / (np.abs(v) + 0.0 ) , -0.5 , 0.5 )
        
        y[0] = slip
        
        return y
    
    
    ###########################################################################
    # For graphical output
    ###########################################################################
    
    
    #############################
    def xut2q( self, x , u , t ):
        """ compute config q """
        
        q   =  x
        
        return q
    
    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """ 
        Compute points p = [x;y;z] positions given config q 
        ----------------------------------------------------
        - points of interest for ploting
        
        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines
        
        """
        
        # Variables
        
        travel   = q[0]
        ang      = q[3]
        
        # constant
        
        r = self.wheel_radius
        l = self.lenght
        
        #base plot form mother class
        lines_pts = LongitudinalFrontWheelDriveCarWithWheelSlipInput.forward_kinematic_lines(self, q)
        
        
        ###########################
        # Wheels
        ###########################
        
        
        angles = np.arange(0,6.4,0.8)
        n      = angles.size
        
        pts = np.zeros((n*2,3))
        
        for i in range(n):
            a = angles[i] - ang
            pts[i*2,:]  = [ r * np.cos(a) , r * np.sin(a) , 0 ]
            pts[i*2+1,:] = [ 0 , 0 , 0 ]

        pts[:,0] = pts[:,0] + travel + l
        pts[:,1] = pts[:,1] + r
        
        
        lines_pts.append( pts )
                
        return lines_pts
    
    
'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    """ MAIN TEST """
    
    sys = LongitudinalFrontWheelDriveCarWithWheelSlipInput()
    
    sys.plot_slip2force()
    
    sys.x0[1]   = 20
    sys.ubar[0] = -0.1
    
    sys.compute_trajectory( 10, 10001, 'euler')
    
    sys.plot_trajectory('xu')
    
    sys.animate_simulation()


# -*- coding: utf-8 -*-

###############################################################################
import numpy as np
import matplotlib.pyplot as plt
###############################################################################
from pyro.dynamic   import rigidbody
from pyro.kinematic import geometry
from pyro.kinematic import drawing
from pyro.analysis  import graphical
###############################################################################

###############################################################################
        
class Boat2D( rigidbody.RigidBody2D ):    
    """
    Simple planar (3 DoF) boat model 
    
    Partialy based on
    'Low-Speed Maneuvering Models for Dynamic Positioning (3 DOFs)'
    see Section 6.7 in
    Fossen, Thor I. Handbook of marine craft hydrodynamics and motion control 
    2nd Editions. John Wiley & Sons, 2021.

    with Equation of motion (in body frame) described by:
    -------------------------------------------------------
    M(q) dv + C(q,v) v + d(q,v) = B(q) u
    dq = N(q) v
    -------------------------------------------------------
    v        :  dim = (3, 1)     : velocity variables ( foward speed, sway speed, yaw rate) in body frame
    q        :  dim = (3, 1)     : position variables ( x , y , theta ) in global frame
    u        :  dim = (2, 1)     : thruster force vector in body frame
    d(q,v)   :  dim = (3, 1)     : state-dependent dissipative forces in body frame
    M(q)     :  dim = (3, 3)     : inertia matrix in body frame
    C(q,v)   :  dim = (3, 3)     : corriolis matrix in body frame
    B(q)     :  dim = (3, 2)     : actuator matrix in body frame
    N(q)     :  dim = (3, 3)     : transformation matrix from body frame to global frame

    with the following hypothesis:
    - The boat is a planar rigid body with 3 DoF ( surge, sway and yaw)
    - The input is a 2D force vector [ F_x , F_y ] applied at a distance l_t behind the CG
    - The boat is subject to linear and quadratic damping
    - The default quadratic hydronamic forces coef. are taken from
        Fossen, Thor I. Handbook of marine craft hydrodynamics and motion control
        2nd Editions. John Wiley & Sons, 2021.
        See 6.7.1 and Fig. 6.11
        A rough fit on experimental data from a tanker
        - TODO: not fully validated
    - The c.g., c.p and body-frame origin are coincident

    """

    ############################
    def __init__(self):
        """ """

        rigidbody.RigidBody2D.__init__( self , force_inputs = 2, other_inputs = 0)

        self.input_label = ['Tx','Ty']
        self.input_units = ['[N]','[N]']

        # Dynamic properties
        self.mass     = 10000.0
        self.inertia  = 10000.0 * 1.0 ** 2
        self.l_t      = 3.0     # Distance between CG and Thrust vector
        
        # Hydrodynamic coefficients

        # linear damping
        self.damping_coef = np.array([ 2000.0, 20000.0, 10000.0 ])

        # quadratic damping
        self.Cx_max = 0.5 
        self.Cy_max = 0.6
        self.Cm_max = 0.1

        self.rho = 1000.0          # water density
        self.Alc = self.l_t * 2    # lateral area
        self.Afc = 0.25 * self.Alc  # frontal area
        self.loa = self.l_t * 2    # length over all

        # current velocity in world frame
        # TODO: not implemented
        # self.v_current = np.array([0,0,0])    

        # Graphic output parameters 
        self.width           = self.Afc
        self.height          = self.l_t * 2
        self.dynamic_domain  = True
        self.dynamic_range   = 10
        
        l   = self.height * 0.5
        w   = self.width  * 0.5
        pts      = np.zeros(( 6 , 3 ))
        pts[0,:] = np.array([-l, +w,0])
        pts[1,:] = np.array([-l, -w,0])
        pts[2,:] = np.array([+l, -w,0])
        pts[3,:] = np.array([l+w,0,0])
        pts[4,:] = np.array([+l, +w,0])
        pts[5,:] = np.array([-l, +w,0])
        
        self.drawing_body_pts = pts

        self.show_hydrodynamic_forces = False

         # State working range
        xy_range = l * 3
        self.x_ub = np.array([+xy_range,+xy_range,+np.pi,10,10,10])
        self.x_lb = np.array([-xy_range,-xy_range,-np.pi,-10,-10,-10])

        # Input working range
        self.u_ub = np.array([+10000,+1000])
        self.u_lb = np.array([-10000,-1000])

    ###########################################################################
    def B(self, q , u ):
        """ 
        Actuator Matrix 
        ------------------
        Here u is a 2D point force [ F_x , F_y ]
        applied at a point located at a distance l_t behind the CG
        hence also creating a yaw moment
        """
        
        B = np.zeros((3,2))
        
        B[0,0] = 1
        B[1,1] = 1
        B[2,1] = - self.l_t 
        
        return B
    
    ###########################################################################
    def CurrentCoef(self, alpha ):
        
        # Cl = np.sin( 2 * alpha )     # flat plate approx
        # Cd = 1 - np.cos( 2 * alpha ) # flat plate approx
        # Cm = 0.0

        # Fig. 7.6 from Fossen  (1st edition)
        # Fig. 6.11 from Fossen (2nd edition)
        Cx = - self.Cx_max * np.cos( alpha  ) * np.abs( np.cos( alpha ) ) 
        Cy = + self.Cy_max * np.sin( alpha  ) * np.abs( np.sin( alpha ) )
        Cm = + self.Cm_max * np.sin( 2 * alpha )
        
        return np.array([ Cx , Cy , Cm ])
        
    ###########################################################################
    def d(self, q , v , u ):
        """ 
        Hydrodynamic dissipative forces
        -----------------------------------

        The model is a combination of linear and quadratic damping

        """

        # linear damping 
        d_lin = v * self.damping_coef

        V2    = v[0]**2 + v[1]**2
        alpha = -np.arctan2( v[1] , v[0] )

        Cx = self.CurrentCoef( alpha )[0]
        Cy = self.CurrentCoef( alpha )[1]
        Cm = self.CurrentCoef( alpha )[2]

        # quadratic forces
        fx = -0.5 * self.rho * self.Afc * Cx * V2
        fy = -0.5 * self.rho * self.Alc * Cy * V2
        mz = -0.5 * self.rho * self.Alc * self.loa * Cm * V2

        d_quad = np.array([ fx , fy , mz ])
        
        return d_quad + d_lin 
    
    
    ###########################################################################
    def forward_kinematic_domain(self, q ):

        l = self.height * 3
        
        x = q[0]
        y = q[1]
        z = 0
        
        if self.dynamic_domain:
        
            domain  = [ ( -l + x , l + x ) ,
                        ( -l + y , l + y ) ,
                        ( -l + z , l + z ) ]#  
        else:
            
            domain  = [ ( -l , l ) ,
                        ( -l , l ) ,
                        ( -l , l ) ]#
                
        return domain
    
    
    ###########################################################################
    def forward_kinematic_lines(self, q ):
        
        lines_pts = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []
        
        ###########################
        #  body
        ###########################
        
        x     = q[0]
        y     = q[1]
        theta = q[2]
        
        W_T_B    = geometry.transformation_matrix_2D( theta , x , y )
        
        pts_B    = self.drawing_body_pts
        pts_W    = drawing.transform_points_2D( W_T_B , pts_B )

        lines_pts.append( pts_W )
        lines_style.append( '-')
        lines_color.append( 'b' )
        
        ###########################
        #  C.G.
        ###########################
        
        pts      = np.zeros(( 1 , 3 ))
        pts[0,:] = np.array([x,y,0])
        
        lines_pts.append( pts )
        lines_style.append( 'o')
        lines_color.append( 'b' )
                
        return lines_pts , lines_style , lines_color
    
    
    ###########################################################################
    def forward_kinematic_lines_plus(self, x , u , t ):
        """ 
        Graphical output showing trust vectors and hydrodynamic forces
        """
        
        lines_pts = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []

        # M per Newton of force
        f2r = 1.0 / self.u_ub[0] * self.height * 0.5
        
        ###########################
        # trust force vector
        ###########################
        
        vx  = u[0] * f2r
        vy  = u[1] * f2r
        offset = -self.l_t
        
        pts_body = drawing.arrow_from_components( vx , vy , x = offset, origin = 'tip'  )    
        W_T_B    = geometry.transformation_matrix_2D( x[2], x[0] , x[1] )
        pts_W    = drawing.transform_points_2D( W_T_B , pts_body )
        
        lines_pts.append( pts_W )
        lines_style.append( '-')
        lines_color.append( 'r' )


        ###########################
        # hydro forces
        ###########################

        if self.show_hydrodynamic_forces:

            q , v = self.x2q( x )

            f = -self.d( q , v , u)

            pts_body = drawing.arrow_from_components( f[0] * f2r , f[1] * f2r )

            pts_W = drawing.transform_points_2D( W_T_B  , pts_body )

            lines_pts.append( pts_W )
            lines_style.append( '--')
            lines_color.append( 'k' )


                
        return lines_pts , lines_style , lines_color
    

    ###########################################################################
    def plot_alpha2Coefs(self, alpha_min = -3.15, alpha_max = 3.15 ):
        
        alphas = np.arange( alpha_min, alpha_max, 0.05 )
        
        n   = alphas.shape[0]
        Cxs = np.zeros((n,1))
        Cys = np.zeros((n,1))
        Cms = np.zeros((n,1))
        
        for i in range(n):
            Cxs[i] = self.CurrentCoef( alphas[i] )[0]
            Cys[i] = self.CurrentCoef( alphas[i] )[1]
            Cms[i] = self.CurrentCoef( alphas[i] )[2]
        
        fig , ax = plt.subplots(3, figsize=graphical.default_figsize,
                                dpi= graphical.default_dpi, frameon=True)
        
        fig.canvas.manager.set_window_title('Aero curve')
        
        ax[0].plot( alphas , Cxs , 'b')
        ax[0].set_ylabel('Cx', fontsize=graphical.default_fontsize)
        ax[0].set_xlabel('alpha', fontsize=graphical.default_fontsize )
        ax[0].tick_params( labelsize = graphical.default_fontsize )
        ax[0].grid(True)
        
        ax[1].plot( alphas , Cys , 'b')
        ax[1].set_ylabel('Cy', fontsize=graphical.default_fontsize)
        ax[1].set_xlabel('alpha', fontsize=graphical.default_fontsize )
        ax[1].tick_params( labelsize = graphical.default_fontsize )
        ax[1].grid(True)
        
        ax[2].plot( alphas , Cms , 'b')
        ax[2].set_ylabel('Cm', fontsize=graphical.default_fontsize)
        ax[2].set_xlabel('alpha', fontsize=graphical.default_fontsize )
        ax[2].tick_params( labelsize = graphical.default_fontsize )
        ax[2].grid(True)
        
        fig.tight_layout()
        fig.canvas.draw()
        
        plt.show()

    
'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    """ MAIN TEST """
    
    sys = Boat2D()

    sys.x0[0] = 0
    sys.x0[1] = 0
    sys.x0[2] = 0

    sys.x0[3] = 2.0
    sys.x0[4] = 0.0
    sys.x0[5] = 0.0
    
    sys.ubar[0] = 10000
    sys.ubar[1] = 1000

    sys.plot_alpha2Coefs()

    #sys.show_hydrodynamic_forces = True

    sys.plot_trajectory('xu')
    sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Mon Nov 12 20:28:17 2018

@author: Alexandre
"""
##############################################################################
import numpy as np
##############################################################################
from pyro.dynamic  import mountaincar
from pyro.planning import discretizer
from pyro.analysis import costfunction
from pyro.planning import dynamicprogramming
##############################################################################

sys  = mountaincar.MountainCar()

sys.x_ub = np.array([+0.2,+2.0])
sys.x_lb = np.array([-1.7,-2.0])

sys.u_ub[0] = +0.2
sys.u_lb[0] = -0.2

# Discrete world 
grid_sys = discretizer.GridDynamicSystem( sys , [201,201] , [11] )

# Cost Function
qcf = costfunction.QuadraticCostFunction.from_sys(sys)

qcf.xbar = np.array([ 0 , 0 ]) # target
qcf.INF  = 30

qcf.R[0,0] = 10.0

qcf.S[0,0] = 10.0
qcf.S[1,1] = 10.0


# DP algo
dp = dynamicprogramming.DynamicProgrammingWithLookUpTable( grid_sys, qcf)

dp.solve_bellman_equation( tol = 0.01)
#dp.solve_bellman_equation( tol = 1 , animate_cost2go = True )
#dp.solve_bellman_equation( tol = 1 , animate_policy = True )

#dp.animate_cost2go( show = False , save = True )
#dp.animate_policy( show = False , save = True )

dp.clean_infeasible_set()
dp.plot_cost2go_3D()
dp.plot_policy()

ctl = dp.get_lookup_table_controller()

# Simulation
cl_sys = ctl + sys
cl_sys.x0   = np.array([-1., 0.])
cl_sys.compute_trajectory( 25, 10001, 'euler')
cl_sys.plot_trajectory('xu')
cl_sys.plot_phase_plane_trajectory()
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Sun May 12 16:34:44 2019

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.control.robotcontrollers import EndEffectorKinematicController
from pyro.dynamic.manipulator import FiveLinkPlanarManipulator
from pyro.dynamic.manipulator import SpeedControlledManipulator
###############################################################################

# Dynamic model :
# Five link robot with internal velocity controllers at the joints
sys  = SpeedControlledManipulator.from_manipulator(FiveLinkPlanarManipulator())

# Controller
ctl       = EndEffectorKinematicController( sys )
ctl.rbar  = np.array([1.0,1.0]) # target effector position
ctl.gains = np.array([2.0,2.0]) # gains
    
# Closed-loop dynamics
cl_sys  = ctl + sys

# Simulation
cl_sys.x0 = np.array([0.1,0.1,0.1,0.1,0.1])  # Initial config
cl_sys.plot_trajectory('x')
cl_sys.animate_simulation()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Oct  3 08:27:06 2021

@author: alex
"""

import numpy as np

from scipy.linalg  import solve_continuous_are

from pyro.dynamic.boat          import Boat2D
from pyro.analysis.costfunction import QuadraticCostFunction
from pyro.dynamic.statespace    import linearize
from pyro.control               import controller

###############################################################################
def na( theta ):
    """ 
    Normalize angle to [-pi,pi]
    """
    
    theta = ( theta + np.pi )  % (2*np.pi) - np.pi
        
    return theta


###############################################################################

class BoatController( controller.StaticController ) :
    """
    A simple boat controller based on a cascade control structure

    High-level position control:
    heading and velocity setpoint are based on the 
    position error computed in the global frame

    Low-level velocity control:
    LQR velocity control computed in the body frame
    
    """
    
    ############################
    def __init__( self , sys ):
        """ """
        # Dimensions of signals
        self.k   = 3 
        self.m   = 2
        self.p   = 6 
        
        super().__init__(self.k, self.m, self.p)
        
        # Label
        self.name = 'Boat Controller'
        
        # Dynamic model available to the controller
        self.sys = sys
        
        # Velocity inner loop parameters
        # only used if the position control loop is deactivated
        self.reference_velocity = np.array([ 5.0, 0.0, 0.0]) 

        # Linearized model for LQR controller
        ss     = linearize( sys , 0.01 )
        self.A = ss.A[3:,3:]
        self.B = ss.B[3:,:]
        self.A[ abs(self.A) < 0.00001 ] = 0.0
        self.B[ abs(self.B) < 0.00001 ] = 0.0
        print('Velocity linearized dynamic')
        print('----------------------------')
        print('A =\n', self.A)
        print('B =\n', self.B)
        
        # Cost function for LQR controller
        cf  = QuadraticCostFunction(3,2)
        cf.Q[0,0] = 10000
        cf.Q[1,1] = 10000
        cf.Q[2,2] = 50000
        cf.R[0,0] = 0.0001
        cf.R[1,1] = 0.0001
        
        # LQR solution
        self.S = solve_continuous_are( self.A , self.B , cf.Q , cf.R )
        self.K = np.linalg.inv( cf.R ) @ self.B.T @ self.S
        self.K[ abs(self.K) < 0.00001 ] = 0.0
        print('Velocity inner-loop LQR gain matrix =\n', self.K)
        
        # Outer position loop parameters
        self.position_control_isactivated = True
        self.KP = np.array([[ 0.5 , 0   , 0],
                            [ 0   , 0.5 , 0],
                            [ 0   , 0   , 2]])
        self.d_max = 2.0

        # Trajectory following parameters
        self.trajectory_following_isactivated = False
        self.traj_amplitude = 12.0
        self.traj_frequency = 0.1

        # Saturations
        self.f_max = np.array([100000,10000])
        self.f_min = np.array([-10000,-10000])
        self.v_max = np.array([5.0,1.0,1.0])
        self.v_min = np.array([-1.0,-1.0,-1.0])

    #############################
    def q_d( self , t = 0 ):
        """ Return the desired position """

        if self.trajectory_following_isactivated:

            a = self.traj_amplitude
            w = self.traj_frequency

            q_d = np.array([ a * np.cos(w*t), 
                             a * np.sin(w*t), 
                             w * t + np.pi/2]) 

        else:

            q_d = np.array([0,0,0.0])

        return q_d
    
    #############################
    def dq_d( self , t = 0 ):
        """ Return the time derivative of the desired position """

        if self.trajectory_following_isactivated:

            a = self.traj_amplitude
            w = self.traj_frequency

            dq_d = np.array([ a * w * - np.sin(w*t) , a * w * np.cos(w * t ), w ])

        else:

            dq_d = np.array([0,0,0.0])

        return dq_d
        
    
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Control law
        """

        # Avilable observations
        q = y[0:3] # position feedback
        v = y[3:]  # velocity feedback

        if self.position_control_isactivated:

            # Desired position and velocity
            q_d  = self.q_d(t)
            dq_d = self.dq_d(t)
            
            # Configuration error
            q_e = q_d - q
            d_e = np.linalg.norm(q_e[0:2])  # absolute distance to target

            # Angular position error withtout cyclic fuck-up
            q_e[2] = na( na(q_d[2]) - na(q[2]) )  

            # Dynamic heading ref
            # If far from target, reference orientation is overided to
            # instead making the boat head toward the desired position
            if d_e > self.d_max:
                actual  = na( q[2] )
                desired = na( np.arctan2( q_e[1] , q_e[0] )  )
                q_e[2]  = na( desired - actual )

            # Position outter loop in inertial frame
            dq_r = self.KP @ q_e + dq_d

            # Desired velocity in body frame
            v_d = self.sys.N( q ).T @ dq_r

            # Velocity setpoint limits
            v_d = np.clip( v_d , self.v_min , self.v_max )

        else:

            # Direct Velocity control for debugging
            v_d = self.reference_velocity
        
        # Velocity error
        v_e = v_d - v
        
        # Velocity inner loop
        u = self.K @ v_e
        
        # Max/min propulsion force
        u = np.clip( u , self.f_min , self.f_max )

        return u
    
    #########################################################################
    def forward_kinematic_lines_plus( self, x , u , t ):
        """  
        Graphical output for the controller
        -----------------------------------
        plot the desired boat pose

        x,u,t are the state, input and time of the global closed-loop system

        """

        if self.position_control_isactivated:

            # desired boat pose, from model forward kinematics
            pts, style, color = self.sys.forward_kinematic_lines( self.q_d(t) )

            # Change the line style and color
            style[0] = '--'
            color[0] = 'c'
            color[1] = 'c'

        else:

            pts = None
            style = None
            color = None

        return pts, style, color


# Non-linear model
sys = Boat2D()

# Cascade controller
ctl = BoatController( sys )

# Simulation of Non-linear Boat Model with LQR linear controller
cl_sys = ctl + sys

################################
# Test velocity innerloop
################################

# ctl.position_control_isactivated = False
# ctl.reference_velocity = np.array([ 1.0, 0.1, 0.2])
# cl_sys.x0 = np.array([3,3,1.0,5,0,0])
# cl_sys.compute_trajectory(10)
# cl_sys.plot_trajectory('xu')
# cl_sys.animate_simulation( time_factor_video = 1.0 )


################################
# Test fixed targets position control
################################

ctl.position_control_isactivated = True

cl_sys.x0 = np.array([3,3,1.0,5,0,0])
cl_sys.compute_trajectory(20)
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation( time_factor_video = 1.0 )

cl_sys.x0 = np.array([-20,10,-2.5,0,0,0])
cl_sys.compute_trajectory(30)
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation( time_factor_video = 1.0 )

cl_sys.x0 = np.array([50,50,0.0,0,0,0])
cl_sys.compute_trajectory(80)
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation( time_factor_video = 1.0 )

################################
# Test trajectory following control
################################

ctl.trajectory_following_isactivated = True

cl_sys.x0 = np.array([0,0,0,0,0,0])
cl_sys.compute_trajectory(60)
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation( time_factor_video = 1.0 )


# -*- coding: utf-8 -*-
"""
Created on Wed Jul 12 10:02:12 2017

@author: alxgr
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from matplotlib import cm
from matplotlib.ticker import LinearLocator

import time

from scipy.interpolate import RectBivariateSpline
from scipy.interpolate import RegularGridInterpolator

'''
################################################################################
'''


class GridDynamicSystem:
    
    ############################
    def __init__(self, sys , x_grid_dim = [ 101 , 101 ], u_grid_dim = [ 11 ] , dt = 0.05 , lookup = True ):
        """
        Class of tools for working with a discretize state space

        Parameters
        ----------
        sys : pyro ContinuousDynamicSystem class
              A dynamic system 
            
        x_grid_dim : list of integers
                     The number of discret level for each dimension of the state space 
            
        u_grid_dim : list of integers
                     The number of discret level for each dimension of the input space 
            
        dt : float
             The time step
            
        lookup : bool
                 option of computing look up table of the foward dynamics
                 
        Returns
        -------
        self.x_level : list of array
                        discrete level for each state coordinates on the grid
                        
        self.u_level : list of  array
                        discrete level for each input coordinates on the grid
                        
        self.nodes_n    : int
                          total number of discrete state on the grid
        
        self.actions_n  : int
                          total number of discrete input on the grid
                          
        self.node_id_from_index : n-D array of int
                                  The node ID based on index for each state coordinates
                                  
        self.action_id_from_index : n-D array of int
                                    The action ID based on index for each input coordinates
                                    
        self.state_from_node_id : 2-D array of float
                                  The state value based on node ID 
        
        self.index_from_node_id : 2-D array of int
                                  The indexes based on node ID 
        
        self.input_from_action_id : 2-D array of float
                                    The state value based on action ID 
        
        self.index_from_action_id : 2-D array of int
                                    The indexes based on action ID 
                                    
        more ...
        
        """
        
        # Dynamic system class
        self.sys = sys 
        
        # time discretization
        self.dt    = dt        
        
        # Grid size
        self.x_grid_dim = np.array( x_grid_dim )
        self.u_grid_dim = np.array( u_grid_dim )
        
        # Options
        self.computelookuptable = lookup
        
        # Plot params
        self.fontsize             = 5
        self.figsize              = (4, 3)
        self.dpi                  = 300
        
        # Initialize
        self.compute()  
        
        
    ##############################
    ### Initial Computations
    ##############################
        
    ##############################
    def compute(self):
        """  """

        self.discretize_state_space()
        self.discretize_input_space() 
        
        print('\nGenerating a mesh for:', self.sys.name)
        print('---------------------------------------------------')
        print('State space dimensions:', self.sys.n , ' Input space dimension:', self.sys.m )
        print('Number of nodes:', self.nodes_n , ' Number of actions:', self.actions_n )
        print('Number of node-action pairs:', self.nodes_n * self.actions_n )
        print('---------------------------------------------------')
        
        self.generate_nodes()
        self.generate_actions()
        
        if self.computelookuptable:
            self.compute_xnext_table()
            self.compute_action_set_table()
            #self.compute_nearest_snext_table() # Unused so far and long
            
        
    #############################
    def discretize_state_space(self):
        """ Grid the state space """
                        
        self.x_level  = []
        self.nodes_n  = 1
        
        # linespace for each x-axis and total number of nodes
        for i in range(self.sys.n):
            self.x_level.append(  np.linspace( self.sys.x_lb[i]  , self.sys.x_ub[i]  , self.x_grid_dim[i]  ) )
            self.nodes_n        = self.nodes_n * self.x_grid_dim[i]
        
        # range and step size for each dim
        self.x_range     = self.sys.x_ub - self.sys.x_lb
        self.x_step_size = self.x_range / ( self.x_grid_dim - 1 )
        
    #############################
    def discretize_input_space(self):
        """ Grid the input space """
        
        self.u_level    = []
        self.actions_n  = 1
        
        # linespace for each u-axis and total number of actions
        for i in range(self.sys.m):
            self.u_level.append(  np.linspace( self.sys.u_lb[i]  , self.sys.u_ub[i]  , self.u_grid_dim[i]  ) )
            self.actions_n       = self.actions_n * self.u_grid_dim[i]
        
        # range and step size for each dim
        self.u_range     = self.sys.u_ub - self.sys.u_lb
        self.u_step_size = self.u_range / ( self.u_grid_dim - 1 )
        
        
    ##############################
    def generate_nodes(self):
        """ Compute 1-D list of nodes based on a regular grid discretization """
        
        start_time = time.time()
        print('Computing nodes..  ', end = '')
        
        # n-D grid of node ID
        self.node_id_from_index = np.zeros( self.x_grid_dim , dtype = int )     # grid of node ID
        
        # 1-D List of nodes
        self.state_from_node_id = np.zeros(( self.nodes_n , self.sys.n ), dtype = float )  # Number of nodes x state dimensions
        self.index_from_node_id = np.zeros(( self.nodes_n , self.sys.n ), dtype = int   )  # Number of nodes x state dimensions
        
        # For all state nodes
        node_id = 0
        
        if self.sys.n == 2 :
            
            for i in range(self.x_grid_dim[0]):
                for j in range(self.x_grid_dim[1]):
                    
                    # State
                    x = np.array([ self.x_level[0][i]  ,  self.x_level[1][j] ])
                    
                    # State and grid index based on node id
                    self.state_from_node_id[ node_id , : ] = x
                    self.index_from_node_id[ node_id , : ] = np.array([i,j])
                    
                    # Node # based on index ij
                    self.node_id_from_index[i,j] = node_id
    
                    # Increment node number
                    node_id = node_id + 1
                    
                    
        elif self.sys.n == 3:
            
            for i in range(self.x_grid_dim[0]):
                for j in range(self.x_grid_dim[1]):
                    for k in range(self.x_grid_dim[2]):
                    
                        # State
                        x = np.array([ self.x_level[0][i]  ,  self.x_level[1][j]  , self.x_level[2][k] ])
                        
                        # State and grid index based on node #
                        self.state_from_node_id[ node_id , : ] = x
                        self.index_from_node_id[ node_id , : ] = np.array([i,j,k])
                        
                        # Node # based on index ijk
                        self.node_id_from_index[i,j,k] = node_id
        
                        # Increment node number
                        node_id = node_id + 1
                        
                        
                        
        elif self.sys.n == 4:
            
            for i in range(self.x_grid_dim[0]):
                for j in range(self.x_grid_dim[1]):
                    for k in range(self.x_grid_dim[2]):
                        for l in range(self.x_grid_dim[3]):
                    
                            # State
                            x = np.array([ self.x_level[0][i]  ,  self.x_level[1][j]  , self.x_level[2][k] , self.x_level[3][l]])
                            
                            # State and grid index based on node #
                            self.state_from_node_id[ node_id , : ] = x
                            self.index_from_node_id[ node_id , : ] = np.array([i,j,k,l])
                            
                            # Node # based on index ijkl
                            self.node_id_from_index[i,j,k,l] = node_id
            
                            # Increment node number
                            node_id = node_id + 1
                    
        else:
            
            raise NotImplementedError
            
        # Print update
        computation_time = time.time() - start_time
        print('completed in %4.2f sec'%computation_time)
            
                
    ##############################
    def generate_actions(self):
        """ Compute 1-D list of actions based on a regular grid discretization"""
        
        start_time = time.time()
        print('Computing actions..  ', end = '')
        
        # m-D grid of action ID
        self.action_id_from_index = np.zeros( self.u_grid_dim , dtype = int )     # grid of node ID
        
        # 1-D List of actions
        self.input_from_action_id = np.zeros(( self.actions_n , self.sys.m ), dtype = float )  # Number of actions x inputs dimensions
        self.index_from_action_id = np.zeros(( self.actions_n , self.sys.m ), dtype = int   )  # Number of actions x inputs dimensions
        
        # For all state nodes
        action_id = 0
        
        # Single input
        
        if self.sys.m == 1 :
        
            for k in range(self.u_grid_dim[0]):
                    
                u = np.array([ self.u_level[0][k] ])
                
                # State and grid index based on node #
                self.input_from_action_id[ action_id , : ] = u
                self.index_from_action_id[ action_id , : ] = k
                
                # Action # based on index k
                self.action_id_from_index[k] = action_id
    
                # Increment node number
                action_id = action_id + 1
                
        elif self.sys.m == 2 :
            
            for k in range(self.u_grid_dim[0]):
                for l in range(self.u_grid_dim[1]):
                    
                    u = np.array([ self.u_level[0][k] , self.u_level[1][l] ])
                    
                    # State and grid index based on node #
                    self.input_from_action_id[ action_id , : ] = u
                    self.index_from_action_id[ action_id , : ] = np.array([k,l])
                    
                    # Action # based on index k
                    self.action_id_from_index[k,l] = action_id
        
                    # Increment node number
                    action_id = action_id + 1
        
        else:
            
            raise NotImplementedError
            
        # Print update
        computation_time = time.time() - start_time
        print('completed in %4.2f sec'%computation_time)
            
            
    ##############################
    def compute_action_set_table(self):
        """ Compute a boolen table describing the action set for each node """
        
        start_time = time.time()
        print('Computing action sets..  ', end = '')
            
        # Evaluation lookup tables      
        self.action_isok   = np.zeros( ( self.nodes_n , self.actions_n ) , dtype = bool )
        
        # For all state nodes        
        for node_id in range( self.nodes_n ):  
            
                x = self.state_from_node_id[ node_id , : ]
            
                # For all control actions
                for action_id in range( self.actions_n ):
                    
                    u = self.input_from_action_id[ action_id , : ]

                    u_ok = self.sys.isavalidinput(x,u)

                    self.action_isok[ node_id , action_id ] = u_ok
                    
        computation_time = time.time() - start_time
        print('completed in %4.2f sec'%computation_time)
                    
                    
    ##############################
    def compute_xnext_table(self):
        """ Compute a x_next lookup table for the forward dynamics """
        
        start_time = time.time()
        print('Computing x_next array.. ', end = '')
            
        # Evaluation lookup tables
        self.x_next_table = np.zeros( ( self.nodes_n , self.actions_n , self.sys.n ) , dtype = float ) # lookup table for dynamic
        self.x_next_isok  = np.zeros( ( self.nodes_n , self.actions_n ) , dtype = bool )
        
        # For all state nodes        
        for node_id in range( self.nodes_n ):  
            
                x = self.state_from_node_id[ node_id , : ]
            
                # For all control actions
                for action_id in range( self.actions_n ):
                    
                    u = self.input_from_action_id[ action_id , : ]
                    
                    # Compute next state for all inputs
                    x_next = self.sys.f( x , u ) * self.dt + x
                    x_ok   = self.sys.isavalidstate( x_next )
                    
                    self.x_next_table[ node_id , action_id , : ] = x_next
                    self.x_next_isok[  node_id , action_id ]     = x_ok
                
                
                if (node_id % 10000) == 9999:
                    computation_time = time.time() - start_time
                    print('\rComputing x_next array.. %d/%d nodes computed in %4.2f sec'%((node_id+1), self.nodes_n ,  computation_time ) )
                
        
        computation_time = time.time() - start_time
        print('\rComputing x_next array.. completed in %4.2f sec'%computation_time)
                    
    
    ##############################
    def compute_nearest_snext_table(self):
        """ Compute s_next lookup table for the forward dynamics """
            
        # Evaluation lookup tables
        self.s_next_table = np.zeros( ( self.nodes_n , self.actions_n ) , dtype = int ) # lookup table for dynamic
        
        # For all state nodes        
        for node_id in range( self.nodes_n ):  
            
                x = self.state_from_node_id[ node_id , : ]
            
                # For all control actions
                for action_id in range( self.actions_n ):
                    
                    # Compute the control input
                    u = self.input_from_action_id[ action_id , : ]
                    
                    # Compute next state
                    x_next = self.sys.f( x , u ) * self.dt + x
                    
                    # Compute nearest node
                    s_next = self.get_nearest_node_id_from_state( x_next )
                    
                    # Put in the lookup table
                    self.s_next_table[ node_id ,  action_id ] = s_next
                    
                    
    ##############################
    ### Save load
    ##############################
    
    ##############################
    def save_lookup_tables(self, name = 'grid' ):
        """  
        
        """
        
        np.savez( name, 
                 x_next_table = self.x_next_table ,
                 x_next_isok  = self.x_next_isok ,
                 action_isok  = self.action_isok   )
        
        
        
    ##############################
    def load_lookup_tables(self, name = 'grid' ):
        """  
        
        """
        
        try:
            
            data = np.load( name + '.npz' )
            
        except:
            
            print('\n File not found ')
            
        else:
            
            self.x_next_table = data['x_next_table']
            self.x_next_isok  = data['x_next_isok']
            self.action_isok  = data['action_isok']
        
    
    
                        
    
    ##############################
    ### Quick convertion shorcut 
    ##############################
    
    ##############################
    def get_index_from_state(self, x ):
        """  
        Return state position on the grid in terms of fractionnal indexes 
        """
        
        indexes = np.zeros( self.sys.n , dtype = float )
        
        # for all state dimensions
        for i in range( self.sys.n ):
            
            indexes[i] = ( x[i] - self.sys.x_lb[i] ) / self.x_range[i] * ( self.x_grid_dim[i] - 1 )
        
        return indexes
    
    
    ##############################
    def get_nearest_index_from_state(self, x ):
        """  
        Return nearest indexes on the state-space grid from a state
        """
        
        # Round the indexes to the nearest integer
        nearest_indexes = np.rint( self.get_index_from_state( x ) ).astype(int)
        
        clipped_indexes = np.clip( nearest_indexes , 0 , self.x_grid_dim - 1 )
        
        # SHould we return -1 for out of bounds indexes??
        
        return clipped_indexes
    
    
    ##############################
    def get_nearest_node_id_from_state(self, x ):
        """  
        Return the node id that is the closest on the grid from x
        """
        
        indexes = tuple( self.get_nearest_index_from_state( x ) )
        
        node_id = self.node_id_from_index[ indexes ]
        
        return node_id
    
    
    ##############################
    def get_index_from_input(self, u ):
        """  
        Return sinput position on the grid in terms of fractionnal indexes 
        """
        
        indexes = np.zeros( self.sys.m , dtype = float )
        
        # for all state dimensions
        for i in range( self.sys.m ):
            
            indexes[i] = ( u[i] - self.sys.u_lb[i] ) / self.u_range[i] * ( self.u_grid_dim[i] - 1 )
        
        return indexes
    
    
    ##############################
    def get_nearest_index_from_input(self, u ):
        """  
        Return nearest indexes on the state-space grid from a state
        """
        
        # Round the indexes to the nearest integer
        nearest_indexes = np.rint( self.get_index_from_input( u ) ).astype(int)
        
        clipped_indexes = np.clip( nearest_indexes , 0 , self.u_grid_dim - 1 )
        
        return clipped_indexes
    
    
    ##############################
    def get_nearest_action_id_from_input(self, u ):
        """  
        Return the action id that is the closest on the grid from u
        """
        
        indexes   = tuple( self.get_nearest_index_from_input( u ) )
        
        action_id = self.action_id_from_index[ indexes ]
        
        return action_id
            
    
    ##############################
    ### Tools
    ##############################
    
    ##############################
    def get_grid_from_array(self, J ):
        """  
        convert a scalar value from node_id 1-D array to n-D array (table)
        """
        
        """
        if self.nodes_n != J.size:
            raise ValueError("Grid size does not match optimal action table size")
        
        # n-D grid of values
        J_grid = np.zeros( self.x_grid_dim , dtype = float )
        
        # For all state nodes        
        for node_id in range( self.nodes_n ): 
            
            indexes = tuple( self.index_from_node_id[ node_id , : ] )
            
            J_grid[ indexes ] = J [ node_id ]
            
        return J_grid
        """
        return J.reshape( self.x_grid_dim )
    
    
    ##############################
    def compute_interpolation_function(self, J , method='linear' , bounds_error = True , fill_value = None ):
        """  
        Return interpolation function for value based on x coordinates
        
        Methods: "linear”, “nearest”, “slinear”, “cubic”, and “quintic”
        """
        
        if self.nodes_n != J.size:
            raise ValueError("Grid size does not match data")
        
        # n-D grid of values
        J_grid = self.get_grid_from_array( J )
        
        levels = tuple(self.x_level[i] for i in range(self.sys.n))
        
        interpol = RegularGridInterpolator( levels , J_grid , method , bounds_error , fill_value )
        
        return interpol
    
    
    ##############################
    def compute_bivariatespline_2D_interpolation_function(self, J , kx = 1 , ky = 1 ):
        """  
        Return interpolation function for value based on x coordinates
        
        Methods: "linear”, “nearest”, “slinear”, “cubic”, and “quintic”
        """
        
        if self.sys.n == 2 : 
        
            if self.nodes_n != J.size:
                raise ValueError("Grid size does not match data")
            
            # n-D grid of values
            J_grid = self.get_grid_from_array( J )
            
            interpol = RectBivariateSpline( self.x_level[0] , self.x_level[1] , J_grid , bbox=[None, None, None, None], kx=kx, ky=ky,)
            
        else:
            
            raise NotImplementedError
        
        return interpol
    
    
    ##############################
    def get_input_from_policy(self, pi , k ):
        """  
        from pi array to k coordinate value of the control input
        """
        
        if self.nodes_n != pi.size:
            raise ValueError("Grid size does not match optimal action table size")
            
        uk_array = np.zeros( self.nodes_n , dtype = float )
        
        # For all state nodes        
        for node_id in range( self.nodes_n ): 
            
            a = pi[ node_id ]
            
            uk_array[ node_id ] = self.input_from_action_id[ a , k ]
            
        return uk_array
    
    
    ##############################
    def get_2D_slice_of_grid(self, Z, axis_1 = 0 , axis_2 = 1 ):
        
        if self.sys.n == 2:
            Z_2d = Z
        
        elif self.sys.n > 2:
            
            axis_1_dim = Z.shape[ axis_1 ]
            axis_2_dim = Z.shape[ axis_2 ]
            
            Z_2d = np.zeros( ( axis_1_dim , axis_2_dim ), dtype = float )
            
            # get defaults index for other axis
            indexes = self.get_nearest_index_from_state( self.sys.xbar )
            
            for i in range( axis_1_dim ):
                for j in range( axis_2_dim):
                    
                    indexes[ axis_1 ] = i
                    indexes[ axis_2 ] = j
                    
                    Z_2d[i,j] = Z[ tuple(indexes) ]
                    
        else:
            
            raise NotImplementedError
            
        return Z_2d
    
    
    ##############################
    def plot_grid_value(self, J , name = 'Value on the grid' , x = 0 , y = 1, jmax =  np.inf , jmin = -1 , cmap = 'YlOrRd'):
        """  
        plot a scalar value (array by node-id) on a grid
        
        Parameters
        ----------
        J : n-D numpy array
        
        name : string
               name of the figure
        
        x : int 
            index of J axis to plot as the x-axis on the graph
            
        y : int 
            index of J axis to plot as the y-axis on the graph
            
        jmax : float
               maximum value to clip the J array on the plot
            
        jmin : float
               minimum value to clip the J array on the plot
        """
        
        ##################################
        # Figure init
        ##################################
        
        fig = plt.figure(figsize= self.figsize, dpi=self.dpi, frameon=True)
        fig.canvas.manager.set_window_title( name )
        ax  = fig.add_subplot(1, 1, 1)

        xname = self.sys.state_label[x] + ' ' + self.sys.state_units[x]
        yname = self.sys.state_label[y] + ' ' + self.sys.state_units[y]
        
        ax.set_ylabel(yname, fontsize=self.fontsize)
        ax.set_xlabel(xname, fontsize=self.fontsize)
        
        x_level = self.x_level[ x ]
        y_level = self.x_level[ y ]
        
        ##################################
        ### Create grid of data and plot
        #################################
        
        #J_grid_nd = np.clip( self.get_grid_from_array( J ) , jmin , jmax )
        J_grid_nd = self.get_grid_from_array( J ) 
        
        J_grid_2d = self.get_2D_slice_of_grid( J_grid_nd , x , y )
        
        mesh = ax.pcolormesh( x_level, y_level, J_grid_2d.T, 
                       shading='gouraud' , cmap = cmap) #, norm = colors.LogNorm()

        mesh.set_clim(vmin=jmin, vmax=jmax)
        
        ##################################
        # Figure param
        ##################################
        
        ax.tick_params( labelsize = self.fontsize )
        ax.grid(True)
        
        fig.colorbar( mesh )
        fig.tight_layout()
        fig.show()
        
        return fig, ax, mesh
    
    
    
    ##############################
    def plot_grid_value_3D(self, J , J2 = None , name = 'Value on the grid' , x = 0 , y = 1, jmax =  np.inf , jmin = -1 , cmap = 'YlOrRd'):
        """  
        plot a scalar value (array by node-id) on a grid
        
        Parameters
        ----------
        J : n-D numpy array
        
        name : string
               name of the figure
        
        x : int 
            index of J axis to plot as the x-axis on the graph
            
        y : int 
            index of J axis to plot as the y-axis on the graph
            
        jmax : float
               maximum value to clip the J array on the plot
            
        jmin : float
               minimum value to clip the J array on the plot
        """

        
        ##################################
        # Figure init
        ##################################

        fig = plt.figure(figsize= self.figsize, dpi=self.dpi, frameon=True)
        fig.canvas.manager.set_window_title( name )
        ax = fig.add_subplot(projection='3d')

        xname = self.sys.state_label[x] + ' ' + self.sys.state_units[x]
        yname = self.sys.state_label[y] + ' ' + self.sys.state_units[y]
        
        ax.set_ylabel(yname, fontsize=self.fontsize)
        ax.set_xlabel(xname, fontsize=self.fontsize)
        ax.set_zlabel('J')
        
        x_level = self.x_level[ x ]
        y_level = self.x_level[ y ]
        
        X, Y = np.meshgrid( x_level ,  y_level )
        
        ##################################
        ### Create grid of data and plot
        #################################
        
        #J_grid_nd = np.clip( self.get_grid_from_array( J ) , jmin , jmax )
        J_grid_nd = self.get_grid_from_array( J ) 
        
        J_grid_2d = self.get_2D_slice_of_grid( J_grid_nd , x , y )
        
        surf = ax.plot_surface( X , Y , J_grid_2d.T , cmap = cmap, linewidth=1.0, antialiased=False) #, norm = colors.LogNorm()

        #mesh.set_clim(vmin=jmin, vmax=jmax)
        
        # Option to compare to J2
        if J2 is None:
            pass
        
        else:
            J2_grid_nd = self.get_grid_from_array( J2 ) 
            J2_grid_2d = self.get_2D_slice_of_grid( J2_grid_nd , x , y )
            wire = ax.plot_wireframe( X , Y , J2_grid_2d.T ) 
        
        
        ax.set_zlim( jmin, np.min( [ J.max() , jmax ] ) )
        ax.zaxis.set_major_locator(LinearLocator(10))
        ax.zaxis.set_major_formatter('{x:2.02f}')
        
        ##################################
        # Figure param
        ##################################
        
        ax.tick_params( labelsize = self.fontsize )
        ax.grid(True)
        
        fig.colorbar( surf )
        #fig.tight_layout()
        fig.show()
        
        return fig, ax, surf
                
        
    ##############################
    def plot_control_input_from_policy(self, pi , k , i = 0 , j = 1 ):
        """  
        """
        
        uk_array = self.get_input_from_policy( pi , k )
        
        fig, ax, mesh = self.plot_grid_value( uk_array , self.sys.input_label[ k ] , i , j , self.sys.u_ub[k] , self.sys.u_lb[k] , cmap = 'bwr')
        
        return fig, ax, mesh



'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    """ MAIN TEST """
    
    
    
    
    

    from pyro.dynamic  import pendulum

    sys  = pendulum.SinglePendulum()
    
    G =  GridDynamicSystem( sys )
    
    sys.x_ub = np.array([2.0,2.0])
    sys.x_lb = np.array([-2.0,-2.0])
    sys.u_ub = np.array([1.0,1.0])
    sys.u_lb = np.array([0.0,0.0])
    
    g = GridDynamicSystem( sys , [ 5, 5] , [2] )
    
    



# -*- coding: utf-8 -*-
"""
Created on Sun May 12 16:34:44 2019

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import manipulator
from pyro.control  import nonlinear
###############################################################################

sys = manipulator.ThreeLinkManipulator3D()

ctl  = nonlinear.ComputedTorqueController( sys )

# Target
ctl.rbar = np.array([0,0,0])

closed_loop_robot = ctl + sys
    
closed_loop_robot.x0  = np.array([3.14,-3,2,0,0,0])
closed_loop_robot.plot_trajectory('x')
closed_loop_robot.plot_trajectory('u')

closed_loop_robot.animate_simulation(time_factor_video=1, is_3d=True)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  1 14:08:03 2024

@author: alex
"""

import numpy as np
import matplotlib.pyplot as plt

import gymnasium as gym
from gymnasium import spaces

from pyro.dynamic import pendulum

from stable_baselines3 import PPO
from stable_baselines3.common.env_util import make_vec_env


class SysEnv(gym.Env):

    def __init__(self, sys, dt=0.1, tf=10.0, render_mode=None):

        # x-y ouputs
        y_ub = np.array([+1, +1, sys.x_ub[1]])
        y_lb = np.array([-1, -1, sys.x_lb[1]])

        self.observation_space = spaces.Box(y_lb, y_ub)
        self.action_space = spaces.Box(sys.u_lb, sys.u_ub)

        self.sys = sys
        self.dt = dt

        self.tf = tf
        self.render_mode = render_mode

        # Memory
        self.x = sys.x0
        self.u = sys.ubar
        self.t = 0.0

        if self.render_mode == "human":

            self.animator = self.sys.get_animator()
            self.animator.show_plus(self.x, self.u, self.t)
            plt.pause(0.001)

    def _get_obs(self):

        theta = self.x[0]
        thetadot = self.x[1]

        y = np.array([np.cos(theta), np.sin(theta), thetadot], dtype=np.float32)

        return y

    def _get_info(self):

        return {"state": self.x, "action": self.u}

    def reset(self, seed=None, options=None):

        super().reset(seed=seed)

        self.x = np.random.uniform(np.array([-np.pi, -1]), np.array([np.pi, 1]))
        self.u = self.sys.ubar
        self.t = 0.0

        observation = self._get_obs()
        info = self._get_info()

        return observation, info

    def step(self, u):

        u = np.clip(u, self.sys.u_lb, self.sys.u_ub)
        x = self.x
        t = self.t
        dt = self.dt

        # Derivatives
        dx = self.sys.f(x, u, t)

        # Euler integration
        x_new = x + dx * dt
        t_new = t + dt

        x_new[0] = self.angle_normalize(x_new[0])

        # Sat speed --> I hate they do this in gym env
        if x_new[1] > sys.x_ub[1]:
            x_new[1] = sys.x_ub[1]
        if x_new[1] < sys.x_lb[1]:
            x_new[1] = sys.x_lb[1]

        # Cost function
        r = -self.sys.cost_function.g(x, u, t)

        terminated = False  # t > self.tf

        truncated = t > self.tf  # False #not self.sys.isavalidstate(x_new)

        # Memory update
        self.x = x_new
        self.t = t_new
        self.u = u

        y = self._get_obs()
        info = self._get_info()

        if self.render_mode == "human":
            self._render_frame()

        return y, r, terminated, truncated, info

    def angle_normalize(self, x):

        return ((x + np.pi) % (2 * np.pi)) - np.pi
    
    def render(self):
        if self.render_mode == "human":
            return self._render_frame()

    def _render_frame(self):

        self.animator.show_plus_update(self.x, self.u, self.t)
        plt.pause(0.001)


sys = pendulum.InvertedPendulum()

# Setting physical parameter to reflect the gym environment

# Physical parameters
sys.gravity = 10.0
sys.m1 = 1.0
sys.l1 = 1.0
sys.lc1 = 0.5 * sys.l1
sys.I1 = (1.0 / 12.0) * sys.m1 * sys.l1**2

sys.l_domain = 2 * sys.l1  # graphical domain

# Min/max state and control inputs
sys.x_ub = np.array([+np.pi, +8])
sys.x_lb = np.array([-np.pi, -8])
sys.u_ub = np.array([+2.0])
sys.u_lb = np.array([-2.0])

# Cost Function
# The reward function is defined as: r = -(theta2 + 0.1 * theta_dt2 + 0.001 * torque2)
sys.cost_function.xbar = np.array([0, 0])  # target
sys.cost_function.R[0, 0] = 0.001
sys.cost_function.Q[0, 0] = 1.0
sys.cost_function.Q[1, 1] = 0.1

# Learning
gym_env = SysEnv(sys, dt=0.05, render_mode=None)
model = PPO("MlpPolicy", gym_env, verbose=1)
model.learn(total_timesteps=250000)

# Animation of the pendulum with the learned policy
gym_env = SysEnv(sys, render_mode="human")
y, info = gym_env.reset()
episodes = 10
for episode in range(episodes):
    y, info = gym_env.reset()
    terminated = False
    truncated = False
    print("\n Episode:", episode)
    while not (terminated or truncated):
        u, _states = model.predict(y, deterministic=True)
        y, r, terminated, truncated, info = gym_env.step(u)
        
    


# -*- coding: utf-8 -*-
"""
Created on Fri Aug 07 11:51:55 2015

@author: agirard
"""

import numpy as np

from pyro.analysis import simulation
from pyro.analysis import phaseanalysis
from pyro.analysis import graphical
from pyro.analysis import costfunction

"""
###############################################################################
"""


###############################################################################
class ContinuousDynamicSystem:
    """
    Mother class for continuous dynamical systems

    Main functions
    ---------------------------------------------
    dx = f( x , u , t )    : dynamic equation

    optionnal:
    y = h( x , u , t )     : output equation (by default y = x )
    u = t2u( t )           : time-dependent input signal (by default u = ubar)
    q = xut2q( x , u , t ) : get configuration vector (by default q = x )

    graphic output:
    lines = forward_kinematic_lines( q ) : draw the system based on q

    Signal and Dimentions
    ----------------------------------------------
    x  : state vector             n x 1
    u  : control inputs vector    m x 1
    t  : time                     1 x 1
    y  : output vector            p x 1
    q  : configuration vector     ? x 1   (dimension is not imposed)

    """

    ###########################################################################
    # The two following functions needs to be implemented by child classes
    ###########################################################################

    ############################
    def __init__(self, n=1, m=1, p=1):
        """
        The __init__ method of the Mother class can be used to fill-in default
        labels, units, bounds, and base values.
        """

        #############################
        # Parameters
        #############################

        # Dimensions
        self.n = n
        self.m = m
        self.p = p

        # Labels
        self.name = "ContinuousDynamicSystem"
        self.state_label = []
        self.input_label = []
        self.output_label = []

        # Units
        self.state_units = []
        self.input_units = []
        self.output_units = []

        # Default Label and units
        for i in range(n):
            self.state_label.append("State " + str(i))
            self.state_units.append("")
        for i in range(m):
            self.input_label.append("Input " + str(i))
            self.input_units.append("")
        for i in range(p):
            self.output_label.append("Output " + str(i))
            self.output_units.append("")

        # Default state and input domain
        self.x_ub = np.zeros(self.n) + 10  # States Upper Bounds
        self.x_lb = np.zeros(self.n) - 10  # States Lower Bounds
        self.u_ub = np.zeros(self.m) + 1  # Control Upper Bounds
        self.u_lb = np.zeros(self.m) - 1  # Control Lower Bounds

        # Default state and inputs values
        self.xbar = np.zeros(self.n)
        self.ubar = np.zeros(self.m)
        self.tbar = 0

        # Plot params
        self.domain = [(-10, 10), (-10, 10), (-10, 10)]
        self.linestyle = "o-"
        self.linestyle_plus = "--"
        self.linescolor = "b"
        self.linescolor_plus = "r"
        self.lines_plus = True  # Bool to active second graphic outpout
        self.is_3d = False  # Use 2d plot by default

        ################################
        # Variables
        ################################

        # Initial value for simulations
        self.x0 = np.zeros(self.n)

        # Result of last simulation
        self.traj = None

        # Cost function for evaluation
        # default is a quadratic cost function with diag Q and R matrices
        self.cost_function = costfunction.QuadraticCostFunction.from_sys(self)

    #############################
    def f(self, x, u, t):
        """
        Continuous time foward dynamics evaluation dx = f(x,u,t)

        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1

        OUPUTS
        dx : state derivative vector  n x 1

        """

        dx = np.zeros(self.n)  # State derivative vector

        ################################################
        # Place holder: put the equations of motion here
        raise NotImplementedError
        ################################################

        return dx

    ###########################################################################
    # The following functions can be overloaded when necessary by child classes
    ###########################################################################

    #############################
    def h(self, x, u, t):
        """
        Output fonction y = h(x,u,t)

        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1

        OUTPUTS
        y  : output vector            p x 1

        """

        # y = np.zeros(self.p) # Output vector

        y = x  # default output is all states

        return y

    #############################
    def t2u(self, t):
        """
        Reference signal fonction u = t2u(t)

        INPUTS
        t  : time                     1 x 1

        OUTPUTS
        u  : control inputs vector    m x 1

        Defaul is a constant signal equal to self.ubar, can be overloaded
        with a more complexe reference signal time-function

        """

        # Default is a constant signal
        u = self.ubar

        return u

    ###########################################################################
    # Basic domain checks, overload if something more complex is needed
    ###########################################################################

    #############################
    def isavalidstate(self, x):
        """check if x is in the state domain"""
        ans = False
        for i in range(self.n):
            ans = ans or (x[i] < self.x_lb[i])
            ans = ans or (x[i] > self.x_ub[i])

        return not (ans)

    #############################
    def isavalidinput(self, x, u):
        """check if u is in the control inputs domain given x"""
        ans = False
        for i in range(self.m):
            ans = ans or (u[i] < self.u_lb[i])
            ans = ans or (u[i] > self.u_ub[i])

        return not (ans)

    ###########################################################################
    # Place holder graphical output, overload with specific graph output
    ###########################################################################

    #############################
    def xut2q(self, x, u, t):
        """Compute configuration variables ( q vector )"""

        # default is q = x

        return x

    ###########################################################################
    def forward_kinematic_domain(self, q):
        """Set the domain range for ploting, can be static or dynamic"""

        return self.domain

    ###########################################################################
    def forward_kinematic_lines(self, q):
        """
        Compute points p = [x;y;z] positions given config q
        ----------------------------------------------------
        - points of interest for ploting

        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines

        """

        lines_pts = []  # list of array (n_pts x 3) for each lines

        ###########################
        # Your graphical code here
        ###########################

        # simple place holder
        for q_i in q:
            pts = np.zeros((1, 3))  # array of 1 pts for the line
            pts[0, 0] = q_i  # x cord of point 0 = q
            lines_pts.append(pts)  # list of all arrays of pts

        return lines_pts

    ###########################################################################
    def forward_kinematic_lines_plus(self, x, u, t):
        """
        Additionnal optionnal graphic output used in animations

        Compute points p = [x;y;z] positions given state x , u , t
        -------------------------------------------------------------
        - additionnal points of interest for ploting
        - for instances forces arrows illustrating control inputs

        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines

        """

        lines_pts = []  # list of array (n_pts x 3) for each lines

        ###########################
        # Your graphical code here
        ###########################

        # simple place holder
        for u_i in u:
            pts = np.zeros((1, 3))  # array of 1 pts for the line
            pts[0, 0] = u_i  # x cord of point 0 = u
            lines_pts.append(pts)  # list of all arrays of pts

        return lines_pts

    ###########################################################################
    # No need to overwrite the following functions for custom dynamic systems
    ###########################################################################

    #############################
    def fsim(self, x, t=0):
        """
        Continuous time foward dynamics evaluation dx = f(x,t), inlcuding the
        internal reference input signal computation

        INPUTS
        x  : state vector             n x 1
        t  : time                     1 x 1

        OUPUTS
        dx : state derivative vector  n x 1

        """

        u = self.t2u(t)
        dx = self.f(x, u, t)

        return dx

    #############################
    def x_next(self, x, u, t=0, dt=0.1, steps=1):
        """
        Discrete time foward dynamics evaluation
        -------------------------------------
        - using Euler integration

        """

        x_next = np.zeros(self.n)  # k+1 State vector

        # Multiple integration steps
        for i in range(steps):

            x_next = self.f(x, u, t) * dt + x

            # Multiple steps
            x = x_next

        return x_next

    ###########################################################################
    # Quick Analysis Shorcuts
    ###########################################################################
    """
    #############################
    def set_single_output_to(self, o=0 ):
        
        #TODO
       
        self.p    = 1             # output size
        
        self.full_h = self.h
    
        # new output fonction
        def new_h(x, u, t):
            
            # New output function
            y_old = self.full_h( x, u, t)
            
            y_new    = np.zeros(1)
            y_new[0] = y_old[o]
            
            return y_new
            
        self.h = new_h
        
        
        pass
    """

    #############################
    def get_plotter(self):
        """Return a Plotter object with param based on sys instance"""

        return graphical.TrajectoryPlotter(self)

    #############################
    def get_animator(self):
        """Return an Animator object with param based on sys instance"""

        return graphical.Animator(self)

    #############################
    def plot_phase_plane(self, x_axis=0, y_axis=1):
        """
        Plot Phase Plane vector field of the system
        ------------------------------------------------
        x_axis : index of state on x axis
        y_axis : index of state on y axis

        """

        pp = phaseanalysis.PhasePlot(self, x_axis, y_axis)

        pp.plot()

    #############################
    def compute_trajectory(self, tf=10, n=10001, solver="solve_ivt", **solver_args):
        """
        Simulation of time evolution of the system
        ------------------------------------------------
        tf : final time
        n  : time steps
        """

        sim = simulation.Simulator(self, tf, n, solver)

        # solve and save the result in the instance
        self.traj = sim.compute(**solver_args)

        return self.traj

    #############################
    def plot_trajectory(self, plot="x", **kwargs):
        """
        Plot time evolution of a simulation of this system
        ------------------------------------------------
        note: will call compute_trajectory if no simulation data is present

        """

        # Check if trajectory is already computed
        if self.traj == None:
            self.compute_trajectory()

        plotter = self.get_plotter()

        return plotter.plot(self.traj, plot, **kwargs)

    #############################
    def plot_phase_plane_trajectory(self, x_axis=0, y_axis=1):
        """
        Plot a trajectory in the Phase Plane
        ---------------------------------------------------------------
        note: will call compute_trajectory if no simulation data is present

        """

        # Check is trajectory is already computed
        if self.traj == None:
            self.compute_trajectory()

        plotter = self.get_plotter()

        return plotter.phase_plane_trajectory(self.traj, x_axis, y_axis)

    #############################
    def plot_phase_plane_trajectory_3d(self, x_axis=0, y_axis=1, z_axis=2):
        """
        Plot the trajectory in the Phase Plane
        ---------------------------------------------------------------
        note: will call compute_trajectory if no simulation data is present

        """

        # Check is trajectory is already computed
        if self.traj == None:
            self.compute_trajectory()

        plotter = self.get_plotter()

        return plotter.phase_plane_trajectory_3d(self.traj, x_axis, y_axis, z_axis)

    #############################################
    def show(self, q, x_axis=0, y_axis=1):
        """Plot figure of configuration q"""

        ani = self.get_animator()
        ani.x_axis = x_axis
        ani.y_axis = y_axis

        ani.show(q)

    #############################################
    def show3(self, q):
        """Plot figure of configuration q"""

        ani = self.get_animator()

        ani.show3(q)

    ##############################
    def animate_simulation(self, **kwargs):
        """
        Show Animation of the simulation
        ----------------------------------
        time_factor_video < 1 --> Slow motion video
        is_3d = True for 3D animation
        note: will call compute_trajectory if no simulation data is present

        """

        # Check is trajectory is already computed
        if self.traj == None:
            self.compute_trajectory()

        ani = self.get_animator()

        return ani.animate_simulation(self.traj, **kwargs)

    ##############################
    def generate_simulation_html_video(self, **kwargs):
        """
        Show Animation of the simulation
        ----------------------------------
        time_factor_video < 1 --> Slow motion video
        is_3d = True for 3D animation
        note: will call compute_trajectory if no simulation data is present

        """

        # Check is trajectory is already computed
        if self.traj == None:
            self.compute_trajectory()

        animator = self.get_animator()
        animator.animate_simulation(self.traj, show=False, **kwargs)
        html_video = animator.ani.to_html5_video()

        return html_video

    #############################
    def generate_mode_animation_html(self, i=0, is_3d=False):
        """
        Linearize and show eigen mode i with html output
        """

        from pyro.dynamic.statespace import linearize

        #
        ss = linearize(self)

        # Compute eigen decomposition
        ss.compute_eigen_modes()

        # Simulate one mode
        traj = ss.compute_eigen_mode_traj(i)

        # Animate mode
        animator = ss.get_animator()

        # label
        template = "Mode %i \n%0.1f+%0.1fj"
        label = template % (i, ss.poles[i].real, ss.poles[i].imag)
        animator.top_right_label = label

        animator.animate_simulation(traj, 3.0, is_3d, show=False)

        html_video = animator.ani.to_html5_video()

        return html_video

    #############################
    def plot_linearized_bode(self, u_index=0, y_index=0):
        """
        Bode plot of linearized siso

        """

        from pyro.dynamic.statespace import linearize
        from pyro.dynamic.tranferfunction import ss2tf

        linearized_sys = linearize(self)
        siso_sys = ss2tf(linearized_sys, u_index, y_index)
        siso_sys.bode_plot()

    #############################
    def plot_linearized_pz_map(self, u_index=0, y_index=0):
        """ """

        from pyro.dynamic.statespace import linearize
        from pyro.dynamic.tranferfunction import ss2tf

        linearized_sys = linearize(self)
        siso_sys = ss2tf(linearized_sys, u_index, y_index)
        siso_sys.pz_map()

    #############################
    def animate_linearized_mode(self, i=0):
        """
        Linearize and show eigen mode i
        """

        from pyro.dynamic.statespace import linearize

        linearized_sys = linearize(self)

        linearized_sys.animate_eigen_mode(i, self.is_3d)

        return linearized_sys

    #############################
    def animate_linearized_modes(self):
        """
        Linearize and show eigen modes

        """

        from pyro.dynamic.statespace import linearize

        linearized_sys = linearize(self)

        animations = []

        for i in range(self.n):
            ani = linearized_sys.animate_eigen_mode(i, self.is_3d)

            animations.append(ani)

        return linearized_sys, animations

    #############################
    def convert_to_gymnasium(self, dt=0.05, tf=10.0, t0=0.0, render_mode=None):
        """
        Create a gym environment from the system

        """

        try:

            import gymnasium as gym
            from gymnasium import spaces
            from pyro.tools.sys2gym import Sys2Gym

            gym_sys = Sys2Gym(self, dt, tf, t0, render_mode)

            return gym_sys

        except:

            raise ImportError("gym library is not installed")


"""
#################################################################
##################          Main                         ########
#################################################################
"""


if __name__ == "__main__":
    """MAIN TEST"""

    pass



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 12 20:28:17 2018

@author: Alexandre
"""
##############################################################################
import numpy as np
##############################################################################
from pyro.dynamic  import pendulum
from pyro.planning import randomtree
##############################################################################


# Dynamic sysem
sys  = pendulum.SinglePendulum()

# Define planning problem
x_start = np.array([0.1,0])
x_goal  = np.array([-3.14,0])

planner = randomtree.RRT( sys , x_start )

planner.u_options = [
            np.array([-5]),
            np.array([-3]),
            np.array([-1]),
            np.array([ 0]),
            np.array([ 1]),
            np.array([ 3]),
            np.array([ 5])
            ]

planner.goal_radius = 0.2

# Solve Planning Problem
planner.find_path_to_goal( x_goal )

# Print solution
#planner.save_solution('pendulum_rrt.npy')
planner.plot_tree()
planner.show_solution()
planner.animate_solution()



# -*- coding: utf-8 -*-
"""
Created on Sun Mar  6 15:09:12 2016

@author: alex
"""

###############################################################################
import scipy.signal as signal           
            


###############################################################################
class TrajectoryFilter:
    """ Low pass filter function for discrete data array """
    
    ############################
    def __init__(self, fc = 5 , dt = 0.01 ):
        """
        Low Pass filter
        
        fc : cutoff   freq [Hz]
        fs : sampling fred [Hz]
        dt : sampling period [sec]
        
        ------------------
        use:
        
        filtered_data = low_pass.filter_array( data )
        
        """
        
        self.fc    = fc
        self.fs    = 1.0 / dt
        self.order = 2
        
        self.setup()
        
    ############################
    def setup(self):
        """ compute intermediate parametes """
        
        wn = 2 * self.fc / self.fs
        
        self.b , self.a  = signal.butter( self.order, wn, btype = 'low' )
        
        
    ############################
    def set_freq_to(self, fc = 10 , dt = 0.01 ):
        """ Recomputed filter given a new cutoff fred """
        
        self.fc    = fc
        self.fs    = 1.0 / dt
        self.setup()
        
        
    ############################
    def low_pass_filter_traj(self, traj ):
        
        filtered_traj = traj.copy()
        
        filtered_traj.x  = signal.filtfilt( self.b, self.a, traj.x.T ).T
        filtered_traj.dx = signal.filtfilt( self.b, self.a, traj.dx.T ).T
        filtered_traj.y  = signal.filtfilt( self.b, self.a, traj.y.T ).T
        filtered_traj.u  = signal.filtfilt( self.b, self.a, traj.u.T ).T
        
        return filtered_traj
        


# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 12:05:08 2018

@author: Alexandre
"""
###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import pendulum
from pyro.control  import linear
###############################################################################


class SinglePendulum_with_position_output( pendulum.SinglePendulum ):
    
    def __init__(self):
        super().__init__()
        
        self.p    = 1             # output size
        #self.rbar = np.array([0]) # ref size


    def h(self, x, u, t):
        
        # New output function
        y = pendulum.SinglePendulum.h(self, x, u, t)
        
        y_position    = np.zeros(1)
        y_position[0] = y[0]
        
        return y_position

sys = SinglePendulum_with_position_output()
dof = 1

kp = 2 # 2,4
kd = 1 # 1
ki = 1

ctl = linear.PIDController(kp, ki, kd)

# Set Point
q_target = np.array([3.14])
ctl.rbar = q_target

# New cl-dynamic
cl_sys = ctl + sys

# Simultation
cl_sys.x0[0] = 1.0
cl_sys.compute_trajectory(tf=10, n=20001, solver='euler')
cl_sys.plot_phase_plane_trajectory()
cl_sys.plot_trajectory('xu')
cl_sys.plot_internal_controller_states()
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 12:05:08 2018

@author: Alexandre
"""
###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic.pendulum      import SinglePendulum
from pyro.analysis.costfunction import QuadraticCostFunction
from pyro.control.lqr           import linearize_and_synthesize_lqr_controller
###############################################################################


sys = SinglePendulum()

# Nominal state to stabilize
sys.xbar[0] = -3.14

# Cost function
cf  = QuadraticCostFunction.from_sys( sys )

cf.Q[0,0] = 1
cf.Q[1,1] = 1

# LQR controller
ctl = linearize_and_synthesize_lqr_controller( sys , cf )

# Simulation Closed-Loop Non-linear with LQR controller
cl_sys = ctl + sys
cl_sys.x0 = np.array([0,0])
cl_sys.compute_trajectory()
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Jun 2 2021

@author: Alex
"""
from pyro.dynamic               import massspringdamper
from pyro.control.linear        import PIDController


# Plant
sys = massspringdamper.FloatingTwoMass( output_mass = 1 )

sys.m1 = 1
sys.m2 = 1
sys.k2 = 3
sys.b2 = 1

sys.compute_ABCD()


# LQR controller
kp  = 2
ki  = 0
kd  = 0
tau = 0.01

ctl = PIDController(kp, ki, kd, tau)

ctl.rbar[0] = 2 # reference

# Simulation Closed-Loop Non-linear with LQR controller
cl_sys = ctl + sys

cl_sys.compute_trajectory(30)
cl_sys.plot_trajectory('y')
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Mon Nov 12 20:28:17 2018

@author: Alexandre
"""
##############################################################################
import numpy as np
##############################################################################
from pyro.dynamic  import pendulum
from pyro.analysis import costfunction
from pyro.planning import dynamicprogramming 
from pyro.planning import discretizer
from pyro.control  import lqr
##############################################################################

sys  = pendulum.SinglePendulum()

sys.xbar  = np.array([ -3.14 , 0 ]) # target and linearization point

##############################
# Cost Function
##############################

qcf     = costfunction.QuadraticCostFunction.from_sys( sys )
qcf.INF = 300

##############################
# VI algo 
##############################

grid_sys = discretizer.GridDynamicSystem( sys , [101,101] , [11] )

dp = dynamicprogramming.DynamicProgrammingWithLookUpTable( grid_sys, qcf)

dp.solve_bellman_equation()
#dp.plot_policy(0)

dp_ctl = dp.get_lookup_table_controller()

##############################
#LQR
##############################

lqr_ctl = lqr.linearize_and_synthesize_lqr_controller( sys, qcf )

##############################
# Policy plots
##############################

dp_ctl.plot_control_law(0,1,0,0,100,sys)
lqr_ctl.plot_control_law(0,1,0,0,100,sys)

##############################
# Simulations
##############################

x0 = np.array([-0 ,0])

cl_sys_lqr =   lqr_ctl + sys 

cl_sys_lqr.x0   = x0
cl_sys_lqr.plot_trajectory('xuj')
cl_sys_lqr.animate_simulation()

cl_sys_vi =   dp_ctl + sys 

cl_sys_vi.x0   = x0
cl_sys_vi.plot_trajectory('xuj')
cl_sys_vi.animate_simulation()



# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 10:17:36 2018

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import integrator
from pyro.analysis import costfunction
###################################

# Simple integrator
sys = integrator.TripleIntegrator()

# Simulation params
sys.ubar = np.array([1]) 
sys.x0   = np.array([2,0,0])

# Compute and show results
sys.plot_trajectory('xuj')
sys.plot_phase_plane_trajectory(0, 1)
sys.plot_phase_plane_trajectory_3d()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Oct 16 22:27:47 2022

@author: alex
"""

import numpy as np

from pyro.dynamic  import massspringdamper
from pyro.control  import controller
from pyro.analysis import costfunction
from pyro.planning import dynamicprogramming
from pyro.planning import discretizer

sys  = massspringdamper.FloatingSingleMass()

sys.x_ub[0] =  10.0
sys.x_lb[0] = -10.0
sys.x_lb[1] = -5.0
sys.x_ub[1] = 5.0
sys.u_ub[0] = 5.0
sys.u_lb[0] = -5.0

# Discrete world 
grid_sys = discretizer.GridDynamicSystem( sys , [101,101] , [41] , 0.05)

# Cost Function
qcf = costfunction.QuadraticCostFunction.from_sys(sys)

qcf.xbar = np.array([ -0 , 0 ]) # target
qcf.INF  = 300

qcf.R[0,0] = 10.0

qcf.S[0,0] = 10.0
qcf.S[1,1] = 10.0


# DP algo
dp = dynamicprogramming.DynamicProgrammingWithLookUpTable( grid_sys, qcf)

dp.solve_bellman_equation( tol = 0.5 )

ctl = dp.get_lookup_table_controller()

dp.plot_cost2go()
dp.plot_cost2go_3D()

ctl.plot_control_law( sys = sys , n = 100)


#asign controller
sys.C = np.eye(2)
sys.p = 2
cl_sys = controller.ClosedLoopSystem( sys , ctl )

##############################################################################

# Simulation and animation
cl_sys.x0   = np.array([+5,+3])
cl_sys.compute_trajectory( 20, 10001, 'euler')
cl_sys.plot_trajectory('xu')
cl_sys.plot_phase_plane_trajectory()
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Jun 2 2021

@author: Alex
"""
import numpy as np

from pyro.dynamic               import massspringdamper
from pyro.analysis.costfunction import QuadraticCostFunction
from pyro.control.lqr           import synthesize_lqr_controller


# Plant
sys = massspringdamper.TwoMass()

sys.m1 = 2
sys.m2 = 3
sys.k1 = 5
sys.k2 = 2
sys.b1 = 0.1
sys.b2 = 0.2

sys.compute_ABCD()

#Full state feedback (default of class is x2 output only)
sys.C = np.diag([1,1,1,1])
sys.p = 4 # dim of output vector
sys.output_label = sys.state_label
sys.output_units = sys.state_units

# Cost function
cf  = QuadraticCostFunction.from_sys( sys )
cf.Q[0,0] = 5
cf.Q[1,1] = 10
cf.Q[2,2] = 5
cf.Q[3,3] = 2
cf.R[0,0] = 0.01
sys.cost_function = cf

# LQR controller
ctl = synthesize_lqr_controller( sys , cf )
ctl.ref_label = ['x1d','x2d','dx1d','dx2d']

# Simulation Closed-Loop Non-linear with LQR controller
cl_sys = ctl + sys
cl_sys.x0 = np.array([0,1,0,0])
cl_sys.compute_trajectory()
cl_sys.plot_trajectory('xu')
cl_sys.plot_linearized_pz_map(1,1)
cl_sys.plot_linearized_bode(1,1)
cl_sys.plot_linearized_pz_map(0,0)
cl_sys.plot_linearized_bode(0,0)
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 12:05:08 2018

@author: Alexandre
"""
###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic import pendulum
from adaptive_computed_torque import DoublePendulumAdaptativeController
###############################################################################

sys = pendulum.DoublePendulum()
sys.cost_function = None
ctl = DoublePendulumAdaptativeController( sys )

#Param adapt-control

ctl.z0[0] = 0
ctl.z0[1] = 0
ctl.z0[2] = 0
ctl.z0[3] = 0
ctl.z0[4] = 0

ctl.Kd[0,0] = 5
ctl.Kd[1,1] = 8

ctl.lam = 1.2

ctl.T[0,0] = 12
ctl.T[1,1] = 12
ctl.T[2,2] = 12
ctl.T[3,3] = 12
ctl.T[4,4] = 12

# Set Point
ctl.rbar = np.array([0,0])

# New cl-dynamic
cl_sys = ctl + sys

# Simultation
cl_sys.x0[0]  = 3.14
cl_sys.x0[1]  = 2.0

cl_sys.state_label[4] = 'H1'
cl_sys.state_label[5] = 'H2'
cl_sys.state_label[6] = 'C1'
cl_sys.state_label[6] = 'C2'
cl_sys.state_label[8] = 'g'

cl_sys.compute_trajectory(tf=10, n=20001, solver='euler')
cl_sys.plot_phase_plane_trajectory()
cl_sys.plot_trajectory('xu')
cl_sys.plot_internal_controller_states()
cl_sys.animate_simulation()



# -*- coding: utf-8 -*-
"""
Created on Tue Oct 23 20:45:37 2018

@author: Alexandre
"""


##############################################################################
import numpy as np
##############################################################################
from pyro.dynamic import system
##############################################################################


##############################################################################
#
##############################################################################
        
class KinematicBicyleModel( system.ContinuousDynamicSystem ):
    """ 
    Equations of Motion
    -------------------------
    dx   = V cos ( phi )
    dy   = V sin ( phi )
    dphi = V/l tan ( beta )
    """
    
    ############################
    def __init__(self):
        """ """
    
        # Dimensions
        self.n = 3   
        self.m = 2   
        self.p = 3
        
        # initialize standard params
        system.ContinuousDynamicSystem.__init__(self, self.n, self.m, self.p)
        
        # Labels
        self.name = 'Kinematic Bicyle Model'
        self.state_label = ['x','y','theta']
        self.input_label = ['v', 'beta']
        self.output_label = ['x','y','theta']
        
        # Units
        self.state_units = ['[m]','[m]','[rad]']
        self.input_units = ['[m/sec]', '[rad]']
        self.output_units = ['[m]','[m]','[rad]']
        
        # State working range
        self.x_ub = np.array([+5,+2,+3.14])
        self.x_lb = np.array([-5,-2,-3.14])
        
        # Model param
        self.lenght = 1
        
        # Graphic output parameters 
        self.dynamic_domain  = True
        self.dynamic_range   = 10
        
    #############################
    def f(self, x = np.zeros(3) , u = np.zeros(2) , t = 0 ):
        """ 
        Continuous time foward dynamics evaluation
        
        dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUPUTS
        dx : state derivative vectror n x 1
        
        """
        
        dx = np.zeros(self.n) # State derivative vector

        dx[0] = u[0] * np.cos( x[2] )
        dx[1] = u[0] * np.sin( x[2] )
        dx[2] = u[0] * np.tan( u[1] ) * ( 1. / self.lenght) 
        
        return dx
    
    
    ###########################################################################
    # For graphical output
    ###########################################################################
    
    
    #############################
    def xut2q( self, x , u , t ):
        """ compute config q """
        
        q   = np.append(  x , u[1] ) # steering angle is part of the config
        
        return q
    
    
    ###########################################################################
    def forward_kinematic_domain(self, q ):
        """ 
        """
        l = self.dynamic_range
        
        x = q[0]
        y = q[1]
        z = 0
        
        if self.dynamic_domain:
        
            domain  = [ ( -l + x , l + x ) ,
                        ( -l + y , l + y ) ,
                        ( -l + z , l + z ) ]#  
        else:
            
            domain  = [ ( -l , l ) ,
                        ( -l , l ) ,
                        ( -l , l ) ]#
            
                
        return domain
    
    
    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """ 
        Compute points p = [x;y;z] positions given config q 
        ----------------------------------------------------
        - points of interest for ploting
        
        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines
        
        """
        
        lines_pts   = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []
        
        ###########################
        # Top line
        ###########################
            
        pts = np.zeros((2,3))
        
        pts[0,0] = -10000
        pts[0,1] = 1
        pts[1,0] = 10000
        pts[1,1] = 1
        
        lines_pts.append( pts )
        lines_style.append('--')
        lines_color.append('k')
        
        ###########################
        # bottom line
        ###########################
        
        pts = np.zeros((2,3))
        
        pts[0,0] = -10000
        pts[0,1] = -1
        pts[1,0] = 10000
        pts[1,1] = -1
        
        lines_pts.append( pts )
        lines_style.append('--')
        lines_color.append('k')
        
        ###########################
        # Vehicule
        ###########################
        
        pts = np.zeros((3,3))
        
        pts[0,0] = q[0]
        pts[0,1] = q[1]
        pts[1,0] = q[0] + self.lenght * np.cos( q[2] )
        pts[1,1] = q[1] + self.lenght * np.sin( q[2] )
        pts[2,0] = ( q[0] + self.lenght * np.cos( q[2] ) + 
                       0.2 * self.lenght * np.cos( q[2] + q[3] ) )
        pts[2,1] = ( q[1] + self.lenght * np.sin( q[2] ) + 
                       0.2 * self.lenght * np.sin( q[2] + q[3] ) )
        
        lines_pts.append( pts )
        lines_style.append('o-')
        lines_color.append('b')
                
        return lines_pts, lines_style , lines_color
    
    

##############################################################################
# 
##############################################################################
        
class HolonomicMobileRobot( system.ContinuousDynamicSystem ):
    """ 
    Holonomic 2D point-robot
    -----------------------------------
    dx   = u[0]
    dy   = u[1]
    
    """
    
    ############################
    def __init__(self):
        """ """
    
        # Dimensions
        self.n = 2   
        self.m = 2   
        self.p = 2
        
        # initialize standard params
        system.ContinuousDynamicSystem.__init__(self, self.n, self.m, self.p)
        
        # Labels
        self.name = 'Holonomic Mobile Robot'
        self.state_label = ['x','y']
        self.input_label = ['vx', 'vy']
        self.output_label = ['x','y']
        
        # Units
        self.state_units = ['[m]','[m]']
        self.input_units = ['[m/sec]','[m/sec]']
        self.output_units = ['[m]','[m]']
        
        # State working range
        self.x_ub = np.array([ 10, 10])
        self.x_lb = np.array([-10,-10])
        
    #############################
    def f(self, x = np.zeros(3) , u = np.zeros(2) , t = 0 ):
        """ 
        Continuous time foward dynamics evaluation
        
        dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUPUTS
        dx : state derivative vectror n x 1
        
        """
        
        dx = np.zeros(self.n) # State derivative vector

        dx[0] = u[0]
        dx[1] = u[1] 
        
        return dx
    
    
    ###########################################################################
    # For graphical output
    ###########################################################################
    
    #############################
    def xut2q( self, x , u , t ):
        """ compute config q """
        
        q = x # kinematic model : state = config space
        
        return q
    
    
    ###########################################################################
    def forward_kinematic_domain(self, q ):
        """ 
        """
        l = 10
        
        domain  = [ ( -l , l ) ,
                    ( -l , l ) ,
                    ( -l , l ) ]#
            
                
        return domain
    
    
    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """ 
        Compute points p = [x;y;z] positions given config q 
        ----------------------------------------------------
        - points of interest for ploting
        
        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines
        
        """
        
        lines_pts   = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []
        
        ###########################
        # Top line
        ###########################
            
        pts = np.zeros((5,3))
        
        d = 0.2
        
        pts[0,0] = q[0]+d
        pts[0,1] = q[1]+d
        pts[1,0] = q[0]+d
        pts[1,1] = q[1]-d
        pts[2,0] = q[0]-d
        pts[2,1] = q[1]-d
        pts[3,0] = q[0]-d
        pts[3,1] = q[1]+d
        pts[4,0] = q[0]+d
        pts[4,1] = q[1]+d
        
        lines_pts.append( pts )
        lines_style.append('-')
        lines_color.append('b')
                
        return lines_pts , lines_style , lines_color
        
        

##############################################################################
#
##############################################################################

class HolonomicMobileRobotwithObstacles( HolonomicMobileRobot ):
    """
    Holonomic 2D point robot with obstacles with allowable domain
    -------------------------------------------------------------
    dx   = u[0]
    dy   = u[1]

    """

    ############################
    def __init__(self):
        """ """
        # initialize standard params
        HolonomicMobileRobot.__init__(self)
        
        # Labels
        self.name = 'Holonomic Mobile Robot with Obstacles'

        # State working range
        self.x_ub = np.array([ 10, 10])
        self.x_lb = np.array([-10,-10])

        self.obstacles = [
                [ (2,2),(4,10)],
                [ (6,-8),(8,8)],
                [ (-8,-8),(-1,8)]
                ]

    #############################
    def isavalidstate(self , x ):
        """ check if x is in the state domain """

        ans = False

        for i in range(self.n):
            ans = ans or ( x[i] < self.x_lb[i] )
            ans = ans or ( x[i] > self.x_ub[i] )

        for obs in self.obstacles:
            on_obs = (( x[0] > obs[0][0]) and
                      ( x[1] > obs[0][1]) and
                      ( x[0] < obs[1][0]) and
                      ( x[1] < obs[1][1]) )

            ans = ans or on_obs

        return not(ans)


    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """
        Compute points p = [x;y;z] positions given config q
        ----------------------------------------------------
        - points of interest for ploting

        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines

        """

        lines_pts = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []

        ###########################
        # Vehicule
        ###########################

        pts = np.zeros((5,3))

        d = 0.2

        pts[0,0] = q[0]+d
        pts[0,1] = q[1]+d
        pts[1,0] = q[0]+d
        pts[1,1] = q[1]-d
        pts[2,0] = q[0]-d
        pts[2,1] = q[1]-d
        pts[3,0] = q[0]-d
        pts[3,1] = q[1]+d
        pts[4,0] = q[0]+d
        pts[4,1] = q[1]+d

        lines_pts.append( pts )
        lines_style.append('-')
        lines_color.append('b')

        ###########################
        # obstacles
        ###########################

        for obs in self.obstacles:

            pts = np.zeros((5,3))

            pts[0,0] = obs[0][0]
            pts[0,1] = obs[0][1]

            pts[1,0] = obs[0][0]
            pts[1,1] = obs[1][1]

            pts[2,0] = obs[1][0]
            pts[2,1] = obs[1][1]

            pts[3,0] = obs[1][0]
            pts[3,1] = obs[0][1]

            pts[4,0] = obs[0][0]
            pts[4,1] = obs[0][1]

            lines_pts.append( pts )
            lines_style.append('-')
            lines_color.append('k')

        
                
        return lines_pts , lines_style , lines_color


##############################################################################
#
##############################################################################

class Holonomic3DMobileRobot(system.ContinuousDynamicSystem):
    """
    Holonomic 3D point-robot
    -----------------------------------
    dx   = u[0]
    dy   = u[1]

    """

    ############################
    def __init__(self):
        """ """

        # Dimensions
        self.n = 3
        self.m = 3
        self.p = 3

        # initialize standard params
        system.ContinuousDynamicSystem.__init__(self, self.n, self.m, self.p)

        # Labels
        self.name = 'Holonomic 3D Mobile Robot'
        self.state_label = ['x', 'y', 'z']
        self.input_label = ['vx', 'vy', 'vz']
        self.output_label = ['x', 'y', 'z']

        # Units
        self.state_units = ['[m]', '[m]', '[m]']
        self.input_units = ['[m/sec]', '[m/sec]', '[m/sec]']
        self.output_units = ['[m]', '[m]', '[m]']

        # State working range
        self.x_ub = np.array([10, 10, 10])
        self.x_lb = np.array([-10, -10, -10])

    #############################
    def f(self, x=np.zeros(3), u=np.zeros(2), t=0):
        """
        Continuous time foward dynamics evaluation

        dx = f(x,u,t)

        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1

        OUPUTS
        dx : state derivative vectror n x 1

        """

        dx = np.zeros(self.n)  # State derivative vector

        dx[0] = u[0]
        dx[1] = u[1]
        dx[2] = u[2]

        return dx

    ###########################################################################
    # For graphical output
    ###########################################################################

    #############################
    def xut2q(self, x, u, t):
        """ compute config q """

        q = x  # kinematic model : state = config space

        return q

    ###########################################################################
    def forward_kinematic_domain(self, q):
        """
        """
        l = 10

        domain = [(-l, l),
                  (-l, l),
                  (-l, l)]  #

        return domain

    ###########################################################################
    def forward_kinematic_lines(self, q):
        """
        Compute points p = [x;y;z] positions given config q
        ----------------------------------------------------
        - points of interest for ploting

        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines

        """

        lines_pts   = []  # list of array (n_pts x 3) for each lines

        ###########################
        # Top line
        ###########################

        pts = np.zeros((4, 3))

        d = 0.2

        pts[0, 0] = q[0] + d
        pts[0, 1] = q[1] + d
        pts[0, 2] = q[2]
        pts[1, 0] = q[0] + d
        pts[1, 1] = q[1] - d
        pts[1, 2] = q[2]
        pts[2, 0] = q[0] - d
        pts[2, 1] = q[1] - d
        pts[2, 2] = q[2]
        pts[3, 0] = q[0] - d
        pts[3, 1] = q[1] + d
        pts[3, 2] = q[2]

        lines_pts.append(pts)

        return lines_pts

##############################################################################
#
##############################################################################

class Holonomic3DMobileRobotwithObstacles(Holonomic3DMobileRobot):
    """
    Holonomic 3D point-robot with non-allowable states
    -----------------------------------
    dx   = u[0]
    dy   = u[1]

    """

    ############################
    def __init__(self):
        """ """
        # initialize standard params
        Holonomic3DMobileRobot.__init__(self)

        # Labels
        self.name = 'Holonomic 3D Mobile Robot with Obstacles'

        # State working range
        self.x_ub = np.array([ 10, 10, 10])
        self.x_lb = np.array([-10,-10, -10])

        self.obstacles = [
                [ ( 2, 2,-1),( 4,10,1)],
                [ ( 6,-8,-1),( 8,8,1)],
                [ (-8,-8,-1),(-1,8,1)]
                ]

    #############################
    def isavalidstate(self , x ):
        """ check if x is in the state domain """

        ans = False

        for i in range(self.n):
            ans = ans or ( x[i] < self.x_lb[i] )
            ans = ans or ( x[i] > self.x_ub[i] )

        for obs in self.obstacles:
            on_obs = (( x[0] > obs[0][0]) and
                      ( x[1] > obs[0][1]) and
                      ( x[2] > obs[0][2]) and
                      ( x[0] < obs[1][0]) and
                      ( x[1] < obs[1][1]) and
                      ( x[2] < obs[1][2]) )

            ans = ans or on_obs

        return not(ans)


    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """
        Compute points p = [x;y;z] positions given config q
        ----------------------------------------------------
        - points of interest for ploting

        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines

        """

        lines_pts   = [] # list of array (n_pts x 3) for each lines

        ###########################
        # Vehicule
        ###########################

        pts = np.zeros((4,3))

        d = 0.2

        pts[0, 0] = q[0] + d
        pts[0, 1] = q[1] + d
        pts[0, 2] = q[2]
        pts[1, 0] = q[0] + d
        pts[1, 1] = q[1] - d
        pts[1, 2] = q[2]
        pts[2, 0] = q[0] - d
        pts[2, 1] = q[1] - d
        pts[2, 2] = q[2]
        pts[3, 0] = q[0] - d
        pts[3, 1] = q[1] + d
        pts[3, 2] = q[2]


        lines_pts.append( pts )

        ###########################
        # obstacles
        ###########################

        for obs in self.obstacles:

            pts = np.zeros((5,3))
            
            #TODO: need to update plot to draw cubes

            pts[0,0] = obs[0][0]
            pts[0,1] = obs[0][1]
            pts[0,2] = obs[0][2]

            pts[1,0] = obs[0][0]
            pts[1,1] = obs[1][1]
            pts[1,2] = obs[0][2]

            pts[2,0] = obs[1][0]
            pts[2,1] = obs[1][1]
            pts[2,2] = obs[0][2]

            pts[3,0] = obs[1][0]
            pts[3,1] = obs[0][1]
            pts[3,2] = obs[0][2]

            pts[4,0] = obs[0][0]
            pts[4,1] = obs[0][1]
            pts[4,2] = obs[0][2]

            lines_pts.append( pts )


        return lines_pts

    
##############################################################################
#
##############################################################################
        
class KinematicCarModel( KinematicBicyleModel ):
    """ 
    
    Bicycle model of real sized car
    ------------------------------------------------------------------------
    length = 5 meters
    
    Equations of Motion
    -------------------------
    dx   = V cos ( phi )
    dy   = V sin ( phi )
    dphi = V/l tan ( beta )
    
    """
    
    ############################
    def __init__(self):
        """ """
        
        # initialize standard params
        KinematicBicyleModel.__init__( self )
        
        # Model param
        self.width  = 2.00
        self.a      = 2.00
        self.b      = 3.00
        self.lenght = self.a+self.b    
        self.lenght_tire = 0.40
        self.width_tire = 0.15
        
        # Graphic output parameters 
        self.dynamic_domain  = True
        self.dynamic_range   = self.lenght * 2
        
        
    ###########################################################################
    # For graphical output
    ###########################################################################

    
    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """ 
        Compute points p = [x;y;z] positions given config q 
        ----------------------------------------------------
        - points of interest for ploting
        
        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines
        
        """
        
        lines_pts   = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []
        
        ###########################
        # Top line
        ###########################
            
        pts = np.zeros((2,10))
        
        pts[0,0] = -10000
        pts[0,1] = self.width * 2.25
        pts[1,0] = 10000
        pts[1,1] = self.width * 2.25
        
        lines_pts.append( pts )
        lines_style.append('-')
        lines_color.append('k')
        
        ###########################
        # middle line
        ###########################
        
        pts = np.zeros((2,10))
        
        pts[0,0] = -10000
        pts[0,1] = self.width * 0.75
        pts[1,0] = 10000
        pts[1,1] = self.width * 0.75
        
        lines_pts.append( pts )
        lines_style.append('--')
        lines_color.append('k')
        
        ###########################
        # bottom line
        ###########################
        
        pts = np.zeros((2,10))
        
        pts[0,0] = -10000
        pts[0,1] = - self.width * 0.75
        pts[1,0] = 10000
        pts[1,1] = - self.width * 0.75
        
        lines_pts.append( pts )
        lines_style.append('-')
        lines_color.append('k')
        
        ###########################
        # Car
        ###########################
        
        """
        Here is how the car is drawn:
        
        |---------lenght--------|
        |-----a------|
        *                       *   -
            *               *       |
          d2    *   CG  *    d3     |
                    *              width
          d1    *       *    d4     |
           *                *       |
        *                       *   -
        """
        # Distance of the four corners of the car from the mass center
        d1 = np.sqrt(self.a**2+(self.width/2)**2)
        d3 = np.sqrt((self.lenght-self.a)**2+(self.width/2)**2)
        
        # Angles of the four lines of the car       
        theta1 = np.pi+np.arctan(self.width/2/self.a)+q[2]
        theta2 = np.pi-np.arctan(self.width/2/self.a)+q[2]
        theta3 = np.arctan(self.width/2/(self.lenght-self.a))+q[2]
        theta4 = 2*np.pi-np.arctan(self.width/2/(self.lenght-self.a))+q[2]
        
        # Build first line of the car 
        pts = np.zeros((2,10))
        pts[0,0],pts[0,1],pts[1,0],pts[1,1] = self.draw_line(d1,d1,theta1,theta2,q[0],q[1])
        
        lines_pts.append( pts )
        lines_style.append('-')
        lines_color.append('b')
        
        # Build second line of the car 
        pts = np.zeros((2,10))
        pts[0,0],pts[0,1],pts[1,0],pts[1,1] = self.draw_line(d3,d3,theta3,theta4,q[0],q[1])
        px3 = pts[0,0] # Points used to center the wheels
        py3 = pts[0,1]
        px4 = pts[1,0]
        py4 = pts[1,1]
        
        lines_pts.append( pts )
        lines_style.append('-')
        lines_color.append('b')
        
        # Build third line of the car 
        pts = np.zeros((2,10))
        pts[0,0],pts[0,1],pts[1,0],pts[1,1] = self.draw_line(d1,d3,theta2,theta3,q[0],q[1])
        
        lines_pts.append( pts )
        lines_style.append('-')
        lines_color.append('b')

        # Build third fourth of the car 
        pts = np.zeros((2,10))
        pts[0,0],pts[0,1],pts[1,0],pts[1,1] = self.draw_line(d1,d3,theta1,theta4,q[0],q[1])
        
        lines_pts.append( pts )
        lines_style.append('-')
        lines_color.append('b')
        
        ###########################
        # Wheels
        ###########################
        
        # Distance of the four corners of a tire from the center
        d  = np.sqrt((self.lenght_tire/2)**2+(self.width_tire/2)**2)
        
        # Angles of the four lines of a tire               
        steer1 = np.pi+np.arctan(self.width/2/self.a)+q[3]+q[2]
        steer2 = np.pi-np.arctan(self.width/2/self.a)+q[3]+q[2]
        steer3 = np.arctan(self.width/2/(self.lenght-self.a))+q[3]+q[2]
        steer4 = 2*np.pi-np.arctan(self.width/2/(self.lenght-self.a))+q[3]+q[2]
        
        # Build first line of the left tire 
        pts = np.zeros((2,10))
        pts[0,0],pts[0,1],pts[1,0],pts[1,1] = self.draw_line(d,d,steer1,steer2,px3,py3)
        
        lines_pts.append( pts )
        lines_style.append('-')
        lines_color.append('b')
        
        # Build second line of the left tire 
        pts = np.zeros((2,10))
        pts[0,0],pts[0,1],pts[1,0],pts[1,1] = self.draw_line(d,d,steer3,steer4,px3,py3)
        
        lines_pts.append( pts )
        lines_style.append('-')
        lines_color.append('b')

        # Build third line of the left tire 
        pts = np.zeros((2,10))
        pts[0,0],pts[0,1],pts[1,0],pts[1,1] = self.draw_line(d,d,steer2,steer3,px3,py3)
        
        lines_pts.append( pts )
        lines_style.append('-')
        lines_color.append('b')
        
        # Build fourth line of the left tire 
        pts = np.zeros((2,10))
        pts[0,0],pts[0,1],pts[1,0],pts[1,1] = self.draw_line(d,d,steer1,steer4,px3,py3)
        
        lines_pts.append( pts )
        lines_style.append('-')
        lines_color.append('b')
        
        # Build first line of the right tire 
        pts = np.zeros((2,10))
        pts[0,0],pts[0,1],pts[1,0],pts[1,1] = self.draw_line(d,d,steer1,steer2,px4,py4)
        
        lines_pts.append( pts )
        lines_style.append('-')
        lines_color.append('b')
        
        # Build second line of the right tire 
        pts = np.zeros((2,10))
        pts[0,0],pts[0,1],pts[1,0],pts[1,1] = self.draw_line(d,d,steer3,steer4,px4,py4)
        
        lines_pts.append( pts )
        lines_style.append('-')
        lines_color.append('b')
        
        # Build third line of the right tire 
        pts = np.zeros((2,10))
        pts[0,0],pts[0,1],pts[1,0],pts[1,1] = self.draw_line(d,d,steer2,steer3,px4,py4)
        
        lines_pts.append( pts )
        lines_style.append('-')
        lines_color.append('b')

        # Build first line of the right tire 
        pts = np.zeros((2,10))
        pts[0,0],pts[0,1],pts[1,0],pts[1,1] = self.draw_line(d,d,steer1,steer4,px4,py4)
        
        lines_pts.append( pts )
        lines_style.append('-')
        lines_color.append('b')
                
        return lines_pts , lines_style , lines_color
        
    ##########################################################################        
    def draw_line(self, d1, d2, angle1, angle2, x, y):
        
        x1 = x+d1*np.cos(angle1)
        y1 = y+d1*np.sin(angle1)
        x2 = x+d2*np.cos(angle2)
        y2 = y+d2*np.sin(angle2)
            
        return x1,y1,x2,y2


    
##############################################################################       
class KinematicCarModelwithObstacles( KinematicCarModel ):
    """ 
    Bicycle model of real sized car with non-allowable states
    ------------------------------------------------------------------------
    length = 5 meters
    
    Equations of Motion
    -------------------------
    dx   = V cos ( phi )
    dy   = V sin ( phi )
    dphi = V/l tan ( beta )
    
    """
    
    
    ############################
    def __init__(self):
        """ """
        # initialize standard params
        KinematicCarModel.__init__(self)
        
        # Labels
        self.name = 'Kinematic Car Model with Obstacles'

        self.obstacles = [
                [ (-10, -1),(-5, 1)],
                [ (-4, 2),(1, 4)],
                [ (12, -1),(17, 1)]
                ]
        
    #############################
    def isavalidstate(self , x ):
        """ check if x is in the state domain """
        ans = False
        for i in range(self.n):
            ans = ans or ( x[i] < self.x_lb[i] )
            ans = ans or ( x[i] > self.x_ub[i] )
        
        for obs in self.obstacles:
            on_obs = (( x[0] + self.lenght * 0.5 > obs[0][0]) and  
                      ( x[1] + self.width * 0.5  > obs[0][1]) and 
                      ( x[0] - self.lenght * 0.5 < obs[1][0]) and 
                      ( x[1] - self.width * 0.5 < obs[1][1]) )
                     
            ans = ans or on_obs
            
        return not(ans)
        
       
    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """ 
        Compute points p = [x;y;z] positions given config q 
        ----------------------------------------------------
        - points of interest for ploting
        
        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines
        
        """
        
        lines_pts , lines_style , lines_color = KinematicCarModel.forward_kinematic_lines(self, q )
        

        ###########################
        # obstacles
        ###########################
        
        for obs in self.obstacles:
            
            pts = np.zeros((5,3))
            
            pts[0,0] = obs[0][0]
            pts[0,1] = obs[0][1]
            
            pts[1,0] = obs[0][0]
            pts[1,1] = obs[1][1]
            
            pts[2,0] = obs[1][0]
            pts[2,1] = obs[1][1]
            
            pts[3,0] = obs[1][0]
            pts[3,1] = obs[0][1]
            
            pts[4,0] = obs[0][0]
            pts[4,1] = obs[0][1]
            
            lines_pts.append( pts )
            lines_style.append('-')
            lines_color.append('k')
                
        return lines_pts , lines_style , lines_color

    ###########################################################################



##############################################################################

class UdeSRacecar( KinematicCarModelwithObstacles ):
    """

    Bicycle model of a car with the parameter of UdeS-Racecar prototypes

    """

    ############################
    def __init__(self):
        """ """

        # initialize standard params
        KinematicCarModelwithObstacles.__init__( self )

        # Model param
        self.width = 0.17
        self.a = 0.17
        self.b = 0.17
        self.lenght = self.a + self.b
        self.lenght_tire = 0.04
        self.width_tire = 0.015

        # Graphic output parameters
        self.dynamic_domain = True
        self.dynamic_range = self.lenght * 2
        
        # Labels
        self.name = 'Kinematic Bicyle Model of UdeS-Racecar'

        self.obstacles = [
            [(  -1, -0.1), (-0.5, 0.1)],
            [(-0.4,  0.2), ( 0.1, 0.4)],
            [( 1.2, -0.1), ( 1.7, 0.1)]
        ]
        


##############################################################################
#
##############################################################################
        
class ConstantSpeedKinematicCarModel( KinematicCarModel ):
    """ 
    Equations of Motion
    -------------------------
    dx   = V cos ( phi )
    dy   = V sin ( phi )
    dphi = V/l tan ( beta )
    """
    
    ############################
    def __init__(self):
        """ """
    
        # Dimensions
        self.n = 3   
        self.m = 1   
        self.p = 3
        
        # initialize standard params
        system.ContinuousDynamicSystem.__init__(self, self.n, self.m, self.p)
        
        # Labels
        self.name = 'Constant Speed Kinematic Car Model'
        self.state_label = ['x','y','theta']
        self.input_label = ['beta']
        self.output_label = ['x','y','theta']
        
        # Units
        self.state_units = ['[m]','[m]','[rad]']
        self.input_units = ['[rad]']
        self.output_units = ['[m]','[m]','[rad]']
        
        # State working range
        self.x_ub = np.array([+5,+2,+3.14])
        self.x_lb = np.array([-5,-2,-3.14])
        
        # Model param
        self.speed  = 2.00
        self.width  = 2.00
        self.a      = 2.00
        self.b      = 3.00
        self.lenght = self.a+self.b   
        
        self.lenght_tire = 0.60
        self.width_tire  = 0.25
        
        # Graphic output parameters 
        self.dynamic_domain  = True
        self.dynamic_range   = self.lenght * 2
        
        
    #############################
    def f(self, x = np.zeros(3) , u = np.zeros(1) , t = 0 ):
        """ 
        Continuous time foward dynamics evaluation
        
        dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUPUTS
        dx : state derivative vectror n x 1
        
        """
        
        dx = np.zeros(self.n) # State derivative vector

        dx[0] = self.speed * np.cos( x[2] )
        dx[1] = self.speed * np.sin( x[2] )
        dx[2] = self.speed * np.tan( u[0] ) * ( 1. / self.lenght) 
        
        return dx
    
    
    #############################
    def xut2q( self, x , u , t ):
        """ compute config q """
        
        q   = np.append(  x , u[0] ) # steering angle is part of the config
        
        return q
        

'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    """ MAIN TEST """
    
    # sys = KinematicBicyleModel()
    
    # sys.ubar = np.array([2,-0.5])
    # sys.plot_trajectory()
    # sys.animate_simulation()
    
    sys = KinematicCarModelwithObstacles()
    
    sys.ubar = np.array([2,-0.5])
    sys.plot_trajectory()
    sys.animate_simulation()
    
    # sys = UdeSRacecar()
    
    # sys.ubar = np.array([2,-0.5])
    # sys.plot_trajectory()
    # sys.animate_simulation()
    
    # sys = HolonomicMobileRobotwithObstacles()
    
    # sys.ubar = np.array([1,1])
    # sys.plot_trajectory()
    # sys.animate_simulation()
    
    
        


# -*- coding: utf-8 -*-
"""
Created on March 20 2020

@author: Pierre
"""
###############################################################################
import numpy as np
import math
###############################################################################
from wcrt import WCRT 

from pyro.control import controller

##############################################################################

class AdaptativeController_WCRT( controller.DynamicController ):
    """ 
    Adaptative Controller for fully actuated mechanical systems (single pendulum)
    """
    
    
    ############################
    def __init__( self , model , traj = None ):
        """ """
        
        self.name = 'Adaptive controller'
        
        # Params
        self.A = np.zeros(8)
        self.T=np.eye(8)
        self.Kd = np.eye(3)
        self.lam  = 1   # Sliding surface slope
        self.nab  = 0.1 # Min convergence rate
        
        self.model=model
        
        k = model.dof   
        m = model.m
        p = model.p
        l = self.A.shape[0]
        
        super().__init__(k, l, m, p) 
        
        
    ##############################
    def trig(self, q ):
        """ 
        Compute cos and sin usefull in other computation 
        ------------------------------------------------
        
        """
        
        c1  = np.cos( q[0] )
        s1  = np.sin( q[0] )
        c2  = np.cos( q[1] )
        s2  = np.sin( q[1] )
        c3  = np.cos( q[2] )
        s3  = np.sin( q[2] )
        c23 = np.cos( q[2] + q[1] )
        s23 = np.sin( q[2] + q[1] )
        
        return [c1,s1,c2,s2,c3,s3,c23,s23]

        
    ############################
    def adaptative_variables( self , ddq_d , dq_d , q_d , dq , q ):
        """ 
        
        Given desired trajectory and actual state
        
        """        
        q_e   = q  -  q_d
        dq_e  = dq - dq_d
        
        s      = dq_e  + self.lam * q_e
        dq_r   = dq_d  - self.lam * q_e
        ddq_r  = ddq_d - self.lam * dq_e
        
        return [ s , dq_r , ddq_r ]
        
        
    ############################
    def adaptative_torque( self , Y , s , q , t ):
        """ 
        
        Given actual state, compute torque necessarly to guarantee convergence
        
        """
                
        u_computed      = np.dot( Y , self.A  )
        
        u_discontinuous = np.dot(self.Kd,s)
        
        u_tot = u_computed - u_discontinuous
        
        return u_tot
    
                        
    ############################
    def b(self, z, x, q_d, t):
        
        [ q , dq ]     = self.model.x2q( x ) 
        
        ddq_d          =   np.zeros( self.model.dof )
        dq_d           =   np.zeros( self.model.dof )
        
        [ s , dq_r , ddq_r ]  = self.adaptative_variables( ddq_d , dq_d , 
                                                           q_d , dq , q )
        
        [c1,s1,c2,s2,c3,s3,c23,s23] = self.trig( q )
        
        Y = np.zeros((3,8))
        dz = np.zeros(8)
        
        Y[0,0]=ddq_r[0]
        Y[0,1]=ddq_r[1]*(s2+s2*c3)
        Y[0,2]=ddq_r[2]*s23
        Y[0,3]=(dq[2]*s2*s3+dq[1]*c2+dq[1]*c2*c3)*dq_r[1]
        Y[0,4]=(dq[2]+dq[1])*c23*dq_r[2]
        Y[0,5]=0
        Y[0,6]=0
        Y[0,7]=q[0]
        Y[1,0]=ddq_r[0]*(s2+s2*c3)
        Y[1,1]=ddq_r[1]*(1+c3+c3**2)
        Y[1,2]=ddq_r[2]*(c3+c3**2)
        Y[1,3]=(dq[2]*(s3+s3*c3))*dq_r[1]
        Y[1,4]=(dq[2]*(s3+s3*c3)+dq[1]*(s3+s3*c3)+dq[0]*(c2*c3))*dq_r[2]
        Y[1,5]=c2
        Y[1,6]=c23
        Y[1,7]=q[1]
        Y[2,0]=ddq_r[0]*s23
        Y[2,1]=ddq_r[1]*(c3+c3**2)
        Y[2,2]=ddq_r[2]
        Y[2,3]=(dq[1]*(s3+s3*c3)+dq[1]*c2*c3)*dq_r[1]
        Y[2,4]=0
        Y[2,5]=0
        Y[2,6]=c23
        Y[2,7]=q[2]
        
        b = np.dot(Y.T,s)
        dz=-1*np.dot( self.T , b )
            
        return dz
    
    
    ############################
    def c( self , z , x , q_d , t = 0 ):
        """ 
        
        Given desired fixed goal state and actual state, compute torques
        
        """
        
        [ q , dq ]     = self.model.x2q( x )
        
        ddq_d          =   np.zeros( self.model.dof )
        dq_d           =   np.zeros( self.model.dof )
        
        [ s , dq_r , ddq_r ]  = self.adaptative_variables( ddq_d , dq_d , 
                                                           q_d , dq , q )
        [c1,s1,c2,s2,c3,s3,c23,s23] = self.trig( q )
        
        Y = np.zeros((3,8))
                
        Y[0,0]=ddq_r[0]
        Y[0,1]=ddq_r[1]*(s2+s2*c3)
        Y[0,2]=ddq_r[2]*s23
        Y[0,3]=(dq[2]*s2*s3+dq[1]*c2+dq[1]*c2*c3)*dq_r[1]
        Y[0,4]=(dq[2]+dq[1])*c23*dq_r[2]
        Y[0,5]=0
        Y[0,6]=0
        Y[0,7]=q[0]
        Y[1,0]=ddq_r[0]*(s2+s2*c3)
        Y[1,1]=ddq_r[1]*(1+c3+c3**2)
        Y[1,2]=ddq_r[2]*(c3+c3**2)
        Y[1,3]=(dq[2]*(s3+s3*c3))*dq_r[1]
        Y[1,4]=(dq[2]*(s3+s3*c3)+dq[1]*(s3+s3*c3)+dq[0]*(c2*c3))*dq_r[2]
        Y[1,5]=c2
        Y[1,6]=c23
        Y[1,7]=q[1]
        Y[2,0]=ddq_r[0]*s23
        Y[2,1]=ddq_r[1]*(c3+c3**2)
        Y[2,2]=ddq_r[2]
        Y[2,3]=(dq[1]*(s3+s3*c3)+dq[1]*c2*c3)*dq_r[1]
        Y[2,4]=0
        Y[2,5]=0
        Y[2,6]=c23
        Y[2,7]=q[2]
        
        self.A =self.get_z_integral( z )
                         
        u                     = self.adaptative_torque( Y , s  , q , t )
        
        return u
    

    ############################
    def get_z_integral(self, z):
        """ get intergral error internal states """
        
        return z[:self.l]
    


pi = math.pi

sys = WCRT()
sys.cost_function = None
ctl  = AdaptativeController_WCRT(sys)

#Param Wcrt
sys.d1 = 3
sys.d2 = 2
sys.d3 = 2

sys.k1 = 10
sys.k2 = 10
sys.k3 = 10

sys.m1 = 2
sys.m2 = 1
sys.m3 = 1

sys.l1  = 0.5 
sys.l2  = 0.8
sys.l3  = 0.8
sys.lc1 = 0.4
sys.lc2 = 0.5
sys.lc3 = 0.7

#Param adapt-control
ctl.z0[0] = 5
ctl.z0[1] = 5
ctl.z0[2] = 5
ctl.z0[3] = 0
ctl.z0[4] = 5
ctl.z0[5] = 20
ctl.z0[6] = 10
ctl.z0[7] = 0

ctl.Kd[0,0] = 7
ctl.Kd[1,1] = 7
ctl.Kd[2,2] = 7

ctl.lam = 1.5

ctl.T[0,0] = 8
ctl.T[1,1] = 8
ctl.T[2,2] = 8
ctl.T[3,3] = 8
ctl.T[4,4] = 8
ctl.T[5,5] = 8
ctl.T[6,6] = 8
ctl.T[7,7] = 8

#Set Point
ctl.rbar = np.array([0,-pi/4,pi/2])

#New cl-dynamic
cl_sys = ctl + sys
#cl_sys = sys

#Simultation
cl_sys.x0[0]  = pi/3
cl_sys.x0[1]  = 1
cl_sys.x0[2]  = 0

cl_sys.state_label[6] = 'H1'
cl_sys.state_label[7] = 'H2'
cl_sys.state_label[8] = 'H3'
cl_sys.state_label[9] = 'C2'
cl_sys.state_label[10] = 'C3'
cl_sys.state_label[11] = 'g1'
cl_sys.state_label[12] = 'g2'
cl_sys.state_label[13] = 'k'

cl_sys.compute_trajectory(tf=10, n=10001, solver='euler')
cl_sys.plot_trajectory()
cl_sys.plot_internal_controller_states()
cl_sys.animate_simulation(is_3d = True)




# -*- coding: utf-8 -*-
"""
Created on Sun May 12 16:34:44 2019

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import manipulator
from pyro.control  import robotcontrollers
###############################################################################


# Model
torque_controlled_robot   = manipulator.ThreeLinkManipulator3D()


# Target
q_desired = np.array([0,0,0])
r_desired = torque_controlled_robot.forward_kinematic_effector( q_desired )

# Effector PD 

model = torque_controlled_robot

effector_pd      = robotcontrollers.EndEffectorPD( model )
effector_pd.rbar = r_desired
effector_pd.kp   = np.array([100, 100 , 100])
effector_pd.kd   = np.array([ 10,  10 ,  10])

# Closed-loops
robot_with_effector_pd = effector_pd + torque_controlled_robot 

# Simulation
robot_with_effector_pd.x0  = np.array([3.14,-3,2,0,0,0])
robot_with_effector_pd.compute_trajectory()
robot_with_effector_pd.plot_trajectory('x')
robot_with_effector_pd.plot_trajectory('u')
robot_with_effector_pd.animate_simulation(time_factor_video=1, is_3d=True)



###############################################################################
import numpy as np
import matplotlib.pyplot as plt
###############################################################################
from pyro.control import robotcontrollers
from pyro.dynamic import manipulator
from asimov import Asimov
###############################################################################
gain = 1  # Gain du controlleur

asimov = Asimov()  # Asimov
sc_asimov = manipulator.SpeedControlledManipulator.from_manipulator( asimov )

sc_asimov.x0 = np.array([3*np.pi/4, -np.pi/2, np.pi/2])  # Position initiale [q1, q2, q3]

ctl = robotcontrollers.EndEffectorKinematicController(sc_asimov, gain)  # Déclaration du controlleur
ctl.rbar = np.array([0.5, 0.25, 0])  # Cible

closed_loop_robot = ctl + sc_asimov  # Système boucle fermé

closed_loop_robot.plot_trajectory()  # Calcul de la trajectoire
closed_loop_robot.animate_simulation( is_3d=True )


# -*- coding: utf-8 -*-
"""
Created on Fri Aug 07 11:51:55 2015

@author: agirard
"""

import numpy as np
import matplotlib
import matplotlib.pyplot as plt

from pyro.analysis import graphical

# Embed font type in PDF
matplotlib.rcParams['pdf.fonttype'] = 42
matplotlib.rcParams['ps.fonttype']  = 42
        

###############################################################################
# Phase Plot Object for phase plane analysis
###############################################################################
        
class PhasePlot:
    """ 
    Continous dynamic system phase plot 
    ---------------------------------------------------
    x_axis : index of state to display as x axis
    y_axis : index of state to display as y axis
    
    """
    ############################
    def __init__(self, ContinuousDynamicSystem , x_axis = 0 ,  y_axis = 1):
        
        # System
        self.cds  = ContinuousDynamicSystem
        self.f    = self.cds.f      # dynamic function
        self.xbar = np.copy( self.cds.xbar )  # default state
        self.ubar = np.copy( self.cds.ubar ) # default input
        self.t    = 0               # default time
        
        # Grid
        self.x_axis = x_axis  # Index of state to plot on x axis
        self.y_axis = y_axis  # Index of state to plot on y axis
               
        self.x_axis_min = self.cds.x_lb[ self.x_axis ]
        self.x_axis_max = self.cds.x_ub[ self.x_axis ]
        self.y_axis_min = self.cds.x_lb[ self.y_axis ]
        self.y_axis_max = self.cds.x_ub[ self.y_axis ]
        
        self.x_axis_n = 21
        self.y_axis_n = 21 
        
        # Plotting params
        self.color      = 'b'        
        self.figsize    = graphical.default_figsize
        self.dpi        = graphical.default_dpi
        self.linewidth  = 0.005
        self.streamplot = False
        self.arrowstyle = '->'
        self.headlength = 4.5
        self.fontsize   = graphical.default_fontsize
        
        
    ###########################################################################
    def compute_grid(self):
        
        x = np.linspace( self.x_axis_min , self.x_axis_max , self.x_axis_n )
        y = np.linspace( self.y_axis_min , self.y_axis_max , self.y_axis_n )
        
        self.X, self.Y = np.meshgrid( x, y)
            
        
    ###########################################################################
    def compute_vector_field(self):
        
        self.v = np.zeros(self.X.shape)
        self.w = np.zeros(self.Y.shape)
        
        for i in range(self.x_axis_n):
            for j in range(self.y_axis_n):
                
                # Actual states
                x  = np.copy( self.xbar )   # default value for all states
                x[ self.x_axis ] = self.X[i, j]
                x[ self.y_axis ] = self.Y[i, j]
                
                # States derivative open loop
                dx = self.f( x , self.ubar , self.t ) 
                
                # Assign vector components
                self.v[i,j] = dx[ self.x_axis ]
                self.w[i,j] = dx[ self.y_axis ]
                
                       
    ###########################################################################
    def plot_init(self):
        
        
        self.phasefig = plt.figure( figsize = self.figsize , dpi = self.dpi,
                                   frameon=True)
        self.phasefig.canvas.manager.set_window_title('Phase plane of ' + 
                                                self.cds.name )
        
        
    ###########################################################################
    def plot_vector_field(self):
        
        try:
            self.ax = self.phasefig.axes[0]
        except IndexError:
            self.ax = self.phasefig.add_subplot(111, autoscale_on=False )

        if self.streamplot:
            self.ax.streamplot( self.X, self.Y, self.v, self.w, 
                                    color      =self.color,  
                                    linewidth  = self.linewidth, 
                                    arrowstyle = self.arrowstyle, 
                                    arrowsize  = self.headlength )
        else:
            self.ax.quiver( self.X, self.Y, self.v, self.w, 
                                color     = self.color,  
                                linewidth = self.linewidth)
                                #, headlength = self.headlength )
            
        
    ###########################################################################
    def plot_finish(self):
        
        self.ax.set_xlabel(self.cds.state_label[ self.x_axis ] + ' ' + 
                           self.cds.state_units[ self.x_axis ] , 
                           fontsize = self.fontsize)
        self.ax.set_ylabel(self.cds.state_label[ self.y_axis ] + ' ' + 
                           self.cds.state_units[ self.y_axis ] , 
                           fontsize = self.fontsize)
        
        self.ax.set_xlim([ self.x_axis_min , self.x_axis_max ])
        self.ax.set_ylim([ self.y_axis_min , self.y_axis_max ])
        self.ax.tick_params( labelsize = self.fontsize )
        self.ax.grid(True)
        self.phasefig.tight_layout()
        
        
    ###########################################################################
    def plot(self):
        """ Plot phase plane """
        
        self.compute_grid()
        self.plot_init()
        self.compute_vector_field()
        self.plot_vector_field()
        self.plot_finish()
        self.phasefig.show()
        
        
###############################################################################
# 3D Phase Plot Object for phase plane analysis
###############################################################################
        
        
class PhasePlot3( PhasePlot ):
    """ 
    Continous dynamic system phase plot 3D
    ---------------------------------------------------
    x_axis : index of state to display as x axis
    y_axis : index of state to display as y axis
    z_axis : index of state to display as z axis
    
    """
    ###########################################################################
    def __init__(self, ContinuousDynamicSystem, x_axis=0,  y_axis=1, z_axis=2):
        
        super().__init__(ContinuousDynamicSystem, x_axis, y_axis)
        
        # Smaller resolution
        self.x_axis_n   = 5
        self.y_axis_n   = 5
        self.z_axis_n   = 5
        
        # Z axis
        self.z_axis     = z_axis  
        self.z_axis_min = self.cds.x_lb[ self.z_axis ]
        self.z_axis_max = self.cds.x_ub[ self.z_axis ]
        
        # Plotting params
        self.color      = 'r'
        self.dpi        = 200
        self.linewidth  = 0.5
        self.length     = 0.2
        self.arrowstyle = '->'
        self.fontsize   = 6
        
        
    ###########################################################################
    def compute_grid(self):
        
        x = np.linspace( self.x_axis_min , self.x_axis_max , self.x_axis_n )
        y = np.linspace( self.y_axis_min , self.y_axis_max , self.y_axis_n )
        z = np.linspace( self.z_axis_min , self.z_axis_max , self.z_axis_n )
        
        self.X, self.Y, self.Z = np.meshgrid( x, y, z)
        
            
    ###########################################################################
    def compute_vector_field(self):
        
        self.v = np.zeros(self.X.shape)
        self.w = np.zeros(self.Y.shape)
        self.u = np.zeros(self.Z.shape)
        
        for i in range(self.x_axis_n):
            for j in range(self.y_axis_n):
                for k in range(self.z_axis_n):
                
                    # Actual states
                    x  = np.copy( self.xbar )    # default value for all states
                    x[ self.x_axis ] = self.X[i, j, k]
                    x[ self.y_axis ] = self.Y[i, j, k]
                    x[ self.z_axis ] = self.Z[i, j, k]
                    
                    # States derivative open loop
                    dx = self.f( x , self.ubar , self.t ) 
                    
                    # Assign vector components
                    self.v[i,j,k] = dx[ self.x_axis ]
                    self.w[i,j,k] = dx[ self.y_axis ]
                    self.u[i,j,k] = dx[ self.z_axis ]
                    
                       
    ###########################################################################
    def plot_vector_field(self):
        try:
            self.ax = self.phasefig.axes[0]
        except IndexError:
            self.ax = self.phasefig.add_subplot(111, projection='3d')

        self.ax.quiver( self.X, self.Y, self.Z, self.v, self.w, self.u, 
                       color=self.color,  linewidth = self.linewidth,
                       length = self.length)
                       #, headlength = self.headlength, normalize = True )
        
        
    ###########################################################################
    def plot_finish(self):
        
        self.ax.set_xlabel(self.cds.state_label[ self.x_axis ] + ' ' +
        self.cds.state_units[ self.x_axis ] , fontsize = self.fontsize)
        self.ax.set_ylabel(self.cds.state_label[ self.y_axis ] + ' ' +
        self.cds.state_units[ self.y_axis ] , fontsize = self.fontsize)
        self.ax.set_zlabel(self.cds.state_label[ self.z_axis ] + ' ' +
        self.cds.state_units[ self.z_axis ] , fontsize = self.fontsize)
        
        plt.grid(True)
        plt.tight_layout()
        

